#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos.  If not, see <http://www.gnu.org/licenses/>.

""" Argos numerical data inspector.
"""
from __future__ import print_function

# ----IMPORTANT ----
# Do not do any imports here that (indirectly) import any dependencies (PyQt, numpy, etc)
# The browse function is imported by the argos package, which in turn is imported by setup.py.
# If you import (for instance) numpy here, the setup.py will fail if numpy is not installed.
# Therefore, we do all imports from the argos package in the functions here.

import argparse
import logging
import os
import os.path
import sys

from glob import glob

logging.captureWarnings(True)

logger = logging.getLogger('argos')

logging.basicConfig(level='DEBUG', stream=sys.stderr,
                    #format='%(name)35s %(asctime)s %(filename)25s:%(lineno)-4d : %(levelname)-8s: %(message)s')
                    format='%(asctime)s %(filename)25s:%(lineno)-4d : %(levelname)-8s: %(message)s')

# We are not using **kwargs here so IDEs can see which parameters are expected.
def browse(filePatterns=None, *,
           select=None,
           inspectorFullName=None,
           qtStyle=None,
           styleSheet=None,
           settingsFile=None,
           addTestData=False,
           runTestWalk=False):
    """ Opens the main window(s) for the persistent settings and executes the application.

        :param filePatterns: List of file names (or unix glob patterns like *.h5) that will be
            added to the repository. If only one file or directory is given it will be selected
            and expanded at start up.
        :param select: a path of the repository item that will selected and expanded at start up,
            even if more than one file or directory is given on the command line.
        :param inspectorFullName: The full path name of the inspector that will be loaded
        :param qtStyle: name of qtStyle (E.g. fusion).
        :param styleSheet: a path to an optional Qt Cascading Style Sheet.
        :param settingsFile: file with persistent settings. If None a default will be used.
        :param addTestData: if True, some in-memory test data is added to the repository tree.
        :param runTestWalk: if True, all nodes are visited and the program exits.
    """
    # Import in functions. See comments at the top for more details
    from argos.info import EXIT_CODE_RESTART

    # Creates and runs an ArgosApplication in a while loop to 'restart' application when the
    # plugin registry was edited.
    while True:
        logger.info("Starting browse window...")
        argosApplication = createArgosApp(
            filePatterns=filePatterns,
            select=select,
            inspectorFullName=inspectorFullName,
            qtStyle=qtStyle,
            styleSheet=styleSheet,
            settingsFile=settingsFile,
            addTestData=addTestData,
            runTestWalk=runTestWalk
        )

        exitCode = argosApplication.execute()

        logger.info("Argos finished with exit code: {}".format(exitCode))
        if exitCode != EXIT_CODE_RESTART:
            return exitCode
        else:
            logger.info("----- Restart requested. The Qt event loop will be restarted. -----\n")



def createArgosApp(filePatterns=None, *,
                   select=None,
                   inspectorFullName=None,
                   qtStyle=None,
                   styleSheet=None,
                   settingsFile=None,
                   addTestData=False,
                   runTestWalk=False):
    """ Creates an Argos application, including window(s) and loads the data.

       The Argos application will also create a QtApplication. You will have to call
       argosApp.execute() on the returned Argos application to start the Qt event loop.

        Returns:
            The newly created ArgosApplication object.
    """
    # Import in functions. See comments at the top for more details.
    from argos.info import DEBUGGING
    from argos.qt import QtWidgets, QtCore
    from argos.application import ArgosApplication
    from argos.repo.testdata import createArgosTestData
    from argos.utils.dirs import normRealPath
    from argos.widgets.misc import setApplicationQtStyle, setApplicationStyleSheet

    argosApp = ArgosApplication(settingsFile, runTestWalk=runTestWalk)
    argosApp.loadSettings(inspectorFullName)  # TODO: call in constructor?

    try:
        QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)
    except Exception as ex:
        logger.debug("AA_UseHighDpiPixmaps not available in PyQt4: {}".format(ex))

    if qtStyle:
        availableStyles = QtWidgets.QStyleFactory.keys()
        if  qtStyle not in availableStyles:
            logger.warning("Qt style '{}' is not available on this computer. Use one of: {}"
                           .format(qtStyle, availableStyles))
        else:
            setApplicationQtStyle(qtStyle)

    if not os.path.exists(styleSheet):
        msg = "Stylesheet not found: {}".format(styleSheet)
        print(msg, file=sys.stderr)
        logger.warning(msg)
        #sys.exit(2)
    setApplicationStyleSheet(styleSheet)

    # Load data in common repository before windows are created.

    fileNames = [f for fp in filePatterns for f in glob(fp)]

    argosApp.loadFiles(fileNames)

    if addTestData:
        argosApp.repo.insertItem(createArgosTestData())

    if select:
        selectPath = select
        logger.debug("Using selection path from the command line: {!r}".format(selectPath))
    else:
        if len(fileNames) == 1:
            # Using normpath to remove trailing slashes.
            selectPath = os.path.basename(normRealPath(fileNames[0]))
            logger.debug("Selection path is the only file that is given on the command line: {!r}"
                         .format(selectPath))
        else:
            selectPath = None
            logger.debug("Not selecting any files (selection path = {!r})".format(selectPath))

    if selectPath:
        logger.debug("Selection path: {}".format(selectPath))
        for mainWindow in argosApp.mainWindows:
            mainWindow.trySelectRtiByPath(selectPath)

    return argosApp


def printInspectors(settingsFile):
    """ Prints a list of inspectors
    """
    # Imported here so this module can be imported without Qt being installed.
    from argos.application import ArgosApplication

    argosApp = ArgosApplication(settingsFile)
    argosApp.loadSettings(None)

    print("# Argos' registered inspectors")
    for regItem in argosApp.inspectorRegistry.items:
        print(regItem.name)


def main():
    """ Starts Argos main window
    """
    # Import in functions. See comments at the top for more details
    import numpy as np

    from argos.widgets.constants import NUMPY_LINE_WIDTH
    from argos.info import DEBUGGING, PROJECT_NAME, VERSION, EXIT_CODE_RESTART, resource_directory
    from argos.utils.logs import initLogging
    from argos.utils.misc import remove_process_serial_number

    aboutStr = "{} version: {}".format(PROJECT_NAME, VERSION)
    parser = argparse.ArgumentParser(description = aboutStr)

    parser.add_argument('-v', '--version', action = 'store_true',
        help="Prints the program version and exits")

    parser.add_argument('filePatterns', metavar='FILE', nargs='*',
        help="""Files or directories that will be loaded at start up. Accepts unix-like glob
                patterns, even on Windows. E.g.: 'argos *.h5' opens all files with the h5 extension
                in the current directory.""")

    parser.add_argument('-s', '--select', dest='selectPath',
        help="""Full path name of an item in the data tree that will be selected at start-up.
                E.g. 'file/var/fieldname'.""")

    parser.add_argument('-i', '--inspector', dest='inspector',
        help="""The name of the inspector that will be opened at start up. E.g. Table""")

    parser.add_argument('--list-inspectors', dest='list_inspectors', action = 'store_true',
        help="""Prints a list of available inspectors for the -i option, and exits.""")

    cfgGroup = parser.add_argument_group(
        "config options", description="Options related to style and configuration.")

    cfgGroup.add_argument('--qt-style', dest='qtStyle', help='Qt style. E.g.: fusion')

    cfgGroup.add_argument('--qss', dest='styleSheet',
                        help="Name of Qt Style Sheet file. If not set, the Argos default style "
                             "sheet will be used.")

    cfgGroup.add_argument('-c', '--config-file', metavar='FILE', dest='settingsFile',
        help="Configuration file with persistent settings. When using a relative path the settings "
             "file is loaded/saved to the argos settings directory.")

    cfgGroup.add_argument('--log-config', metavar='FILE', dest='logConfigFileName',
                        help='Logging configuration file. If not set a default will be used.')

    cfgGroup.add_argument('-l', '--log-level', dest='log_level', default='',
        help="Log level. If set, only log messages with a level higher or equal than this will be "
             "printed to screen (stderr). Overrides the log level of the StreamHandlers in the "
             "--log-config file. Does not alter the log level of log handlers that write to a "
             "file.",
        choices=('debug', 'info', 'warning', 'error', 'critical'))

    devGroup = parser.add_argument_group(
        "developer options", description="Options that are mainly useful for Argos developers.")

    devGroup.add_argument('-d', '--debugging-mode', dest='debugging', action = 'store_true',
        help="Run Argos in debugging mode. Useful during development.")

    devGroup.add_argument('--add-test-data', dest='addTestData', action = 'store_true',
        help="Adds some in-memory test data. Useful during development.")

    devGroup.add_argument('--run-test-walk', dest='runTestWalk', action = 'store_true',
        help="Walks through all the nodes in the repository tree and exits. "
             "The exit code 1 if any of the nodes failed to display.")

    args = parser.parse_args(remove_process_serial_number(sys.argv[1:]))

    initLogging(args.logConfigFileName, args.log_level)

    if args.version:
        print(aboutStr)
        sys.exit(0)

    if args.list_inspectors:
        printInspectors(args.settingsFile)
        sys.exit(0)

    logger.info("######################################")
    logger.info("####         Starting Argos       ####")
    logger.info("######################################")
    logger.info(aboutStr)

    logger.debug("argv: {}".format(sys.argv))
    logger.debug("Main argos module file: {}".format(__file__))
    logger.debug("PID: {}".format(os.getpid()))

    if DEBUGGING:
        logger.warning("Debugging flag is on!")

    logger.info('Started {} {}'.format(PROJECT_NAME, VERSION))
    logger.info("Python version: {}".format(sys.version).replace('\n', ''))

    # Imported here so this module can be imported without Qt being installed.
    from argos.qt.bindings import QT_API_NAME
    logger.info("Using {} Python Qt bindings".format(QT_API_NAME))

    # The DEBUGGING 'constant' is set before the arguments are parsed by argparse.
    # Sanity check for consistency.
    assert args.debugging == DEBUGGING, ("Inconsistent debugging mode. {} != {}"
                                         .format(args.debugging, DEBUGGING))

    qtStyle = args.qtStyle if args.qtStyle else os.environ.get("QT_STYLE_OVERRIDE", 'Fusion')
    styleSheet = args.styleSheet if args.styleSheet else os.environ.get("ARGOS_STYLE_SHEET", '')

    if not styleSheet:
        styleSheet = os.path.join(resource_directory(), "argos.css")
        logger.debug("Using default style sheet: {}".format(styleSheet))
    else:
        styleSheet = os.path.abspath(styleSheet)

    logger.debug("Setting numpy line width to {} characters".format(NUMPY_LINE_WIDTH))
    np.set_printoptions(linewidth=NUMPY_LINE_WIDTH)

    # Browse will create an ArgosApplication with one MainWindow
    browse(filePatterns = args.filePatterns,
           inspectorFullName=args.inspector,
           select=args.selectPath,
           qtStyle=qtStyle,
           styleSheet=styleSheet,
           settingsFile=args.settingsFile,
           addTestData=args.addTestData,
           runTestWalk=args.runTestWalk)
    logger.info('Done {}'.format(PROJECT_NAME))

if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Repository tree items that are read using JSON
"""

import logging

from argos.external.json_with_comments import parse_json_with_comments_file
from argos.repo.baserti import BaseRti, lengthToSummary
from argos.repo.memoryrtis import _createFromObject
from argos.repo.iconfactory import RtiIconFactory, ICON_COLOR_UNDEF
from argos.utils.cls import is_a_sequence, is_a_mapping, is_an_array, type_name
from argos.utils.misc import pformat

logger = logging.getLogger(__name__)


def replace_lists_by_arrays(dct):
    """ Recursively walks a JSON dictionary and tries to replace"""


class JsonFileRti(BaseRti):
    """ Read JSON data with any comments filtered out.
        See https://github.com/sidneycadot/json_with_comments
    """
    _defaultIconGlyph = RtiIconFactory.FILE

    def __init__(self, nodeName='', fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor. Initializes as an MappingRti with None as underlying dictionary.
        """
        super(JsonFileRti, self).__init__(nodeName=nodeName, fileName=fileName,
                                         iconColor=iconColor)
        self._checkFileExists()
        self._data = None


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return True


    def _openResources(self):
        """ Uses parse_json_with_comments_file to open the underlying file
        """
        self._data = parse_json_with_comments_file(self._fileName)
        logger.info("READ JSON: type = {}".format(type(self._data)))


    def _fetchAllChildren(self):
        """ Adds a child item for each item
        """
        childItems = []
        logger.debug("_fetchAllChildren of {!r} ({}):  {!r}"
                     .format(self, self.iconColor, self.fileName))

        if self.hasChildren():
            if is_a_sequence(self._data):
                for nr, elem in enumerate(self._data):
                    childItems.append(_createFromObject(
                        elem, nodeName="elem-{}".format(nr),
                        iconColor=self.iconColor, fileName=self.fileName))

            elif is_a_mapping(self._data):
                for key, value in self._data.items():
                    childItems.append(_createFromObject(
                        value, nodeName=str(key),
                        iconColor=self.iconColor, fileName=self.fileName))

            elif is_an_array(self._data):
                raise TypeError("Unexpected type in JSON: {}".format(type(self._data)))

            elif isinstance(self._data, bytearray):
                raise TypeError("Unexpected type in JSON: {}".format(type(self._data)))

            else:
                pass  # Data os a scalar so no add no children.

        return childItems


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._data = None


    def _containsScalar(self):
        """ Returns True if the JSON data consists of just a scalar

            Returns False if the file is closed.
        """
        return (self.isOpen and
                not is_a_sequence(self._data) and
                not is_a_mapping(self._data) and
                not is_an_array(self._data) and
                not isinstance(self._data, bytearray))


    @property
    def isSliceable(self):
        """ Returns True because the underlying data can be sliced.
            The scalar will be wrapped in an array with one element so it can be inspected.
        """
        return self._containsScalar()


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).
            The scalar will be wrapped in an array with one element so it can be inspected.
        """
        assert self._containsScalar(), "JSON element is only slicable when using a scalar"
        return self._data


    @property
    def arrayShape(self):
        """ Returns the shape of the wrapper array. Will always be an empty tuple()
        """
        assert self._containsScalar(), "JSON element is only slicable when using a scalar"
        return tuple()


    @property
    def dimensionality(self):
        """ String that describes if the RTI is an array, scalar, field, etc.
        """
        if not self.isOpen:
            return ""
        elif is_a_sequence(self._data):
            return "list"
        elif is_a_mapping(self._data):
            return ""
        elif is_an_array(self._data):
            return "array"
        elif isinstance(self._data, bytearray):
            return "array"
        else:
            return "scalar"

    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """

        if self._containsScalar():
            return type_name(self._data)
        else:
            return ""


    @property
    def summary(self):
        """ Returns a summary of the contents of the RTI.  E.g. 'array 20 x 30' elements.
        """
        if not self.isOpen:
            return ""
        elif is_a_sequence(self._data):
            return lengthToSummary(len(self._data))
        elif is_a_mapping(self._data):
            return ""  # emtpy to be in-line with other group RTIs
        elif is_an_array(self._data):
            raise NotImplementedError("Not (yet) implemented")
        elif isinstance(self._data, bytearray):
            raise NotImplementedError("Not (yet) implemented")
        else:
            return str(self._data)


    def quickLook(self, width: int):
        """ Returns a string representation fof the RTI to use in the Quik Look pane.

            We print all data, even if it is large, since it is already in memory, and it is
            assumed to be quick.
        """
        if not self.isOpen:
            return ""
        else:
            return pformat(self._data, width)



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Repository Tree Items (RTIs) for HDF-5 data.

    It uses the h5py package to open HDF-5 files.
    See http://www.h5py.org/
"""
from __future__ import absolute_import

import logging, os

import numpy as np
import h5py

from argos.repo.iconfactory import RtiIconFactory, ICON_COLOR_UNDEF
from argos.repo.baserti import BaseRti, shapeToSummary
from argos.utils.cls import to_string, check_class, is_an_array
from argos.utils.defs import DIM_TEMPLATE, SUB_DIM_TEMPLATE, CONTIGUOUS
from argos.utils.masks import maskedEqual
from argos.utils.moduleinfo import versionStrToTuple

logger = logging.getLogger(__name__)

H5PY_MAJOR_VERSION = versionStrToTuple(h5py.__version__)[0]
MAX_QUICK_LOOK_SIZE = 1000


def dimNamesFromDataset(h5Dataset, useBaseName=True):
    """ Constructs the dimension names given a h5py dataset.

        First looks in the dataset's dimension scales to see if it refers to another
        dataset. In that case the referred dataset's name is used. If not, the label of the
        dimension scale is used. Finally, if this is empty, the dimension is numbered.
    """
    dimNames = [] # TODO: cache?
    for dimNr, dimScales in enumerate(h5Dataset.dims):
        if len(dimScales) == 0:
            dimNames.append(DIM_TEMPLATE.format(dimNr))
        else:
            dimScaleLabel, dimScaleDataset = dimScales.items()[0]
            path = dimScaleDataset.name
            if path:
                dimName = os.path.basename(path) if useBaseName else path
            elif dimScaleLabel: # This could potentially be long so it's our second choice
                dimName = dimScaleLabel
            else:
                dimName = DIM_TEMPLATE.format(dimNr)

            if len(dimScales) > 1:
                logger.debug("More than one dimension scale found, using the first {}"
                             .format(dimName))

            dimNames.append(dimName)
        # else:
        #     # TODO: multiple scales for this dimension. What to do?
        #     dimNames.append(DIM_TEMPLATE.format(dimNr)) # For now, just number them

    return dimNames


def dataSetType(dtype, dimensionalityString):
    """ Returns a string describing the element type of the dataset.

        Args:
            dtype: numpy dtype of the dataset
            dimensionalityString: e.g. 'array ', 'scalar ', 'empty '
    """
    if dimensionalityString and dimensionalityString[0] != ' ':
        dimensionalityString = ' ' + dimensionalityString

    if dtype.names:
        return "compound{}".format(dimensionalityString)
    else:
        if H5PY_MAJOR_VERSION <= 2:
            # h5py version <= 2.x
            if dtype.metadata and 'vlen' in dtype.metadata:
                vlen_type = dtype.metadata['vlen']
                try:
                    return "<vlen {}>".format(vlen_type.__name__)  # when vlen_type is a type
                except AttributeError: #
                    return "<vlen {}>".format(vlen_type.name)      # when vlen_type is a dtype
        else:
            # h5py version >= 3.x
            stringInfo = h5py.check_string_dtype(dtype)
            if stringInfo is not None:
                # The dimensionality will be placed before the string encoding and length because
                # it's more important.
                if stringInfo.length is None:
                    return "string{} {} variable length"\
                        .format(dimensionalityString, stringInfo.encoding)
                else:
                    return "string{} {} length = {}"\
                        .format(dimensionalityString, stringInfo.encoding, stringInfo.length)

            vlenType = h5py.check_vlen_dtype(dtype)
            if vlenType is not None:
                # Note that, currently, variable length arrays cannot be displayed in any inspector!
                # In general, they cannot be loaded in a numpy array, only in some cases.
                # Therefore, they are hard to handle generically. Even the h5py documentation says
                # it's better to consider using something else if you've got the choice.
                # See: https://docs.h5py.org/en/stable/special.html#arbitrary-vlen-data
                return "{}{} variable length".format(vlenType, dimensionalityString)

    return "{}{}".format(dtype, dimensionalityString)



def dataSetUnit(h5Dataset):
    """ Returns the unit of the h5Dataset by looking in the attributes.

        It searches in the attributes for one of the following keys:
        'unit', 'units', 'Unit', 'Units', 'UNIT', 'UNITS'. If these are not found, the empty
        string is returned.

        Always returns a string
    """
    attributes = h5Dataset.attrs
    if not attributes:
        return '' # a premature optimization :-)

    for key in ('unit', 'units', 'Unit', 'Units', 'UNIT', 'UNITS'):
        if key in attributes:
            # In Python3 the attributes are byte strings so we must decode them
            # This a bug in h5py, see https://github.com/h5py/h5py/issues/379
            return to_string(attributes[key])
    # Not found
    return ''


def attrsToDict(attrs):
    """ Converts attributes to a dictionary.

        In case an attribute can't be read and raises and exception, the attribute value will be
        the error message. This will make the attributes more robust.

        Necessary because, for instance, h5py may not be able to read all attributes
        generated by new versions of the Python netCDF4 library.
        # See: https://github.com/h5py/h5py/issues/719
    """
    result = {}
    for key in attrs.keys():
        try:
            result[key] = attrs[key]
        except Exception as ex:
            logger.warning("Unable to read {} attribute: {}".format(key, ex))
            result[key] = str(ex)
    return result


def dataSetMissingValue(h5Dataset):
    """ Returns the missingData given a HDF-5 dataset

        Looks for one of the following attributes: _FillValue, missing_value, MissingValue,
        missingValue. Returns None if these attributes are not found.

        HDF-EOS and NetCDF files seem to put the attributes in 1-element arrays. So if the
        attribute contains an array of one element, that first element is returned here.
    """
    attributes = h5Dataset.attrs
    if not attributes:
        return None # a premature optimization :-)

    for key in ('missing_value', 'MissingValue', 'missingValue', 'FillValue', '_FillValue'):
        if key in attributes:
            missingDataValue = attributes[key]
            if is_an_array(missingDataValue) and len(missingDataValue) == 1:
                return missingDataValue[0] # In case of HDF-EOS and NetCDF files
            else:
                return missingDataValue
    return None


def _dataSetQuickLook(data, string_info):
    """ Makes a quick look representation of a dataset.

        Args:
            data: a numpy array or scalar
            string_info: h5py.string_dtype information read with h5py.check_string_dtype
    """
    if H5PY_MAJOR_VERSION <= 2:
        return str(data) # Just convert bytes to string in h5py <= 2.x
    else:
        if string_info is None:
            return str(data)  # not a string
        else:
            # Variable length (vlen) string datasets will yield a numpy array with object dtype.
            # First convert to a byte array so that we can decode it.
            bytesArray = np.char.asarray(data)
            return str(np.char.decode(bytesArray, encoding=string_info.encoding))



class H5pyScalarRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a scalar HDF-5 variable.

    """
    _defaultIconGlyph = RtiIconFactory.SCALAR

    def __init__(self, h5Dataset, nodeName='', fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor
        """
        super(H5pyScalarRti, self).__init__(
            nodeName=nodeName, fileName=fileName, iconColor=iconColor)
        check_class(h5Dataset, h5py.Dataset)
        self._h5Dataset = h5Dataset


    def hasChildren(self):
        """ Returns False. Leaf nodes never have children. """
        return False


    @property
    def isSliceable(self):
        """ Returns True because the underlying data can be sliced.
            The scalar will be wrapped in an array with one element so it can be inspected.
        """
        return True


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).
            The scalar will be wrapped in an array with one element so it can be inspected.
        """
        return self._h5Dataset[()]


    @property
    def arrayShape(self):
        """ Returns the shape of the wrapper array. Will always be an empty tuple()
        """
        return tuple()


    @property
    def dimensionality(self):
        """ String that describes if the RTI is an array, scalar, field, etc.
        """
        return "empty" if self._h5Dataset.shape is None else "scalar"


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        return dataSetType(self._h5Dataset.dtype, '')


    @property
    def typeName(self):
        """ String representation of the type. By default, the elementTypeName + dimensionality.
        """
        return dataSetType(self._h5Dataset.dtype, self.dimensionality)


    @property
    def attributes(self):
        """ The attributes dictionary.
        """
        return attrsToDict(self._h5Dataset.attrs)


    @property
    def unit(self):
        """ Returns the unit of the RTI by calling dataSetUnit on the underlying dataset
        """
        return dataSetUnit(self._h5Dataset)


    @property
    def missingDataValue(self):
        """ Returns the value to indicate missing data. None if no missing-data value is specified.
        """
        # For empty datasets, self._h5Dataset.size can be None
        if self._h5Dataset.size is not None and self._h5Dataset.size > MAX_QUICK_LOOK_SIZE:
            return "{} of {}".format(self.typeName, self.summary)
        else:
            return dataSetMissingValue(self._h5Dataset)


    @property
    def summary(self):
        """ Returns a summary of the contents of the RTI. In this case the scalar as a string
        """
        if H5PY_MAJOR_VERSION <= 2:
            return str(self._h5Dataset[()])  # Just convert bytes to string in h5py <= 2.x
        else:
            # In h5py >= 3.x the encoding is known
            string_info = h5py.check_string_dtype(self._h5Dataset.dtype)

            if self._h5Dataset.shape is None:
                return "empty dataset"
            elif string_info is None:
                return str(self._h5Dataset[()])  # not a string
            else:
                return self._h5Dataset[()].decode(string_info.encoding, errors="replace")


    def quickLook(self, width: int):
        """ Returns a string representation fof the RTI to use in the Quik Look pane.

            Override default so bytes are decoded to strings.
        """
        return self.summary



class H5pyFieldRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a field in a structured HDF-5 variable.
    """
    _defaultIconGlyph = RtiIconFactory.FIELD

    def __init__(self, h5Dataset, nodeName, subArray, fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor.
        """
        super(H5pyFieldRti, self).__init__(nodeName, fileName=fileName, iconColor=iconColor)
        check_class(h5Dataset, h5py.Dataset)
        self._h5Dataset = h5Dataset

        self._subArray = subArray # The array that this field contains. Can be h5Dataset itself.
        self._isStructured = bool(self._subArray.dtype.names)


    def hasChildren(self):
        """ Returns False. Field nodes never have children.
        """
        return self._isStructured


    @property
    def attributes(self):
        """ The attributes dictionary.
            Returns the attributes of the variable that contains this field.
        """
        return attrsToDict(self._h5Dataset.attrs)


    @property
    def isSliceable(self):
        """ Returns True because the underlying data can be sliced.
        """
        return True


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).
            Applies the index on the HDF dataset that contain this field and then selects the
            current field. In pseudo-code, it returns: self.ncVar[index][self.nodeName].

            If the field itself contains a sub-array it returns:
                self.dataset[mainArrayIndex][self.nodeName][subArrayIndex]
        """
        return self._subArray[index]


    @property
    def nDims(self):
        """ The number of dimensions of the underlying array
        """
        return len(self.arrayShape) # h5py datasets don't have an ndim property


    @property
    def arrayShape(self):
        """ Returns the shape of the underlying array.
        """
        return self._subArray.shape


    @property
    def chunking(self):
        """ List with chunk sizes if chunked storage is used. Or 'contiguous' for contiguous storage
        """
        chunks = self._h5Dataset.chunks
        return CONTIGUOUS if chunks is None else chunks


    @property
    def dimensionality(self):
        """ String that describes if the RTI is an array, scalar, field, etc.
        """
        return "field"


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        return dataSetType(self._subArray.dtype, '')


    @property
    def typeName(self):
        """ String representation of the type. By default, the elementTypeName + dimensionality.
        """
        return dataSetType(self._subArray.dtype, self.dimensionality)


    @property
    def dimensionNames(self):
        """ Returns a list with the dimension names of the underlying HDF5 variable
        """
        nSubDims = len(self._subArray.shape)
        subArrayDims = [SUB_DIM_TEMPLATE.format(dimNr) for dimNr in range(nSubDims)]
        return dimNamesFromDataset(self._h5Dataset) + subArrayDims


    @property
    def dimensionPaths(self):
        """ Returns a list with the full path names of the dimensions.
        """
        nSubDims = len(self._subArray.shape)
        subArrayDims = [SUB_DIM_TEMPLATE.format(dimNr) for dimNr in range(nSubDims)]
        return dimNamesFromDataset(self._h5Dataset, useBaseName=False) + subArrayDims


    @property
    def unit(self):
        """ Returns the unit of the RTI by calling dataSetUnit on the underlying dataset
        """
        unit = dataSetUnit(self._h5Dataset)
        fieldNames = self._h5Dataset.dtype.names

        # If the missing value attribute is a list with the same length as the number of fields,
        # return the missing value for field that equals the self.nodeName.
        if hasattr(unit, '__len__') and len(unit) == len(fieldNames):
            idx = fieldNames.index(self.nodeName)
            return unit[idx]
        else:
            return unit


    @property
    def missingDataValue(self):
        """ Returns the value to indicate missing data. None if no missing-data value is specified.
        """
        value = dataSetMissingValue(self._h5Dataset)
        fieldNames = self._h5Dataset.dtype.names

        # If the missing value attribute is a list with the same length as the number of fields,
        # return the missing value for field that equals the self.nodeName.
        if hasattr(value, '__len__') and len(value) == len(fieldNames):
            idx = fieldNames.index(self.nodeName)
            return value[idx]
        else:
            return value


    @property
    def summary(self):
        """ Returns a summary of the contents of the RTI.  E.g. 'array 20 x 30' elements.
        """
        return shapeToSummary(self.arrayShape)


    def quickLook(self, width: int):
        """ Returns a string representation fof the RTI to use in the Quik Look pane.
        """
        if self._h5Dataset.size > MAX_QUICK_LOOK_SIZE:
            return "{} of {}".format(self.typeName, self.summary)
        else:
            subIndex = tuple([Ellipsis])
            slicedArray = self._subArray[subIndex]
            data = maskedEqual(slicedArray, self.missingDataValue)

            string_info = h5py.check_string_dtype(self._h5Dataset.dtype)
            return _dataSetQuickLook(data, string_info)


    def _fetchAllChildren(self):
        """ Fetches all fields that this variable contains.
            Only variables with a structured data type can have fields.
        """
        assert self.canFetchChildren(), "canFetchChildren must be True"

        childItems = []
        if self._isStructured:
            for fieldName in self._subArray.dtype.names:
                childItems.append(H5pyFieldRti(
                    self._h5Dataset, fieldName, self._subArray[fieldName],
                    fileName=self.fileName, iconColor=self.iconColor))
        return childItems


class H5pyDatasetRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a HDF5 dataset.

        This includes dimenions scales, which are then displayed with a different icon.
    """
    #_defaultIconGlyph = RtiIconFactory.ARRAY # the iconGlyph property is overridden below

    def __init__(self, h5Dataset, nodeName, fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor
        """
        super(H5pyDatasetRti, self).__init__(nodeName, fileName=fileName, iconColor=iconColor)
        check_class(h5Dataset, h5py.Dataset)
        self._h5Dataset = h5Dataset
        self._isStructured = bool(self._h5Dataset.dtype.names)


    @property
    def iconGlyph(self):
        """ Shows an Array icon for regular datasets but a dimension icon for dimension scales
        """
        try:
            if self._h5Dataset.attrs.get('CLASS', None) == b'DIMENSION_SCALE':
                return RtiIconFactory.DIMENSION
            else:
                return RtiIconFactory.ARRAY
        except OSError as ex:
            logger.warning("Unable to read 'CLASS' attribute. Assuming regular dataset".format(ex))
            return RtiIconFactory.ARRAY


    def hasChildren(self):
        """ Returns True if the variable has a structured type, otherwise returns False.
        """
        return self._isStructured


    @property
    def isSliceable(self):
        """ Returns True because the underlying data can be sliced.
        """
        return True


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).
            Passes the index through to the underlying dataset.
            Converts to a masked array using the missing data value as fill_value
        """
        # Some old HDF5 files return bytes objects when containing string data. Convert to array.
        array = np.array(self._h5Dataset.__getitem__(index))
        return maskedEqual(array, self.missingDataValue)


    @property
    def arrayShape(self):
        """ Returns the shape of the underlying array.
        """
        return self._h5Dataset.shape


    @property
    def chunking(self):
        """ List with chunk sizes if chunked storage is used. Or 'contiguous' for contiguous storage
        """
        chunks = self._h5Dataset.chunks
        return CONTIGUOUS if chunks is None else chunks


    @property
    def dimensionality(self):
        """ String that describes if the RTI is an array, scalar, field, etc.
        """
        return "array"


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        return dataSetType(self._h5Dataset.dtype, '')


    @property
    def typeName(self):
        """ String representation of the type. By default, the elementTypeName + dimensionality.
        """
        return dataSetType(self._h5Dataset.dtype, self.dimensionality)


    @property
    def attributes(self):
        """ The attributes dictionary.
        """
        return attrsToDict(self._h5Dataset.attrs)


    @property
    def dimensionNames(self):
        """ Returns a list with the dimension names of the underlying HDF-5 dataset.
        """
        return dimNamesFromDataset(self._h5Dataset) # TODO: cache?


    @property
    def dimensionPaths(self):
        """ Returns a list with the full path names of the dimensions.
        """
        return dimNamesFromDataset(self._h5Dataset, useBaseName=False)


    @property
    def unit(self):
        """ Returns the unit of the RTI by calling dataSetUnit on the underlying dataset
        """
        return dataSetUnit(self._h5Dataset)


    @property
    def missingDataValue(self):
        """ Returns the value to indicate missing data. None if no missing-data value is specified.
        """
        return dataSetMissingValue(self._h5Dataset)


    @property
    def summary(self):
        """ Returns a summary of the contents of the RTI.  E.g. 'array 20 x 30' elements.
        """
        return shapeToSummary(self.arrayShape)


    def quickLook(self, width: int):
        """ Returns a string representation fof the RTI to use in the Quik Look pane.
        """
        if self._h5Dataset.size > MAX_QUICK_LOOK_SIZE:
            return "{} of {}".format(self.typeName, self.summary)
        else:
            data = maskedEqual(self._h5Dataset[:], self.missingDataValue)
            string_info = h5py.check_string_dtype(self._h5Dataset.dtype)
            return _dataSetQuickLook(data, string_info)


    def _fetchAllChildren(self):
        """ Fetches all fields that this variable contains.
            Only variables with a structured data type can have fields.
        """
        assert self.canFetchChildren(), "canFetchChildren must be True"

        childItems = []

        # Add fields
        if self._isStructured:
            for fieldName in self._h5Dataset.dtype.names:
                childItems.append(H5pyFieldRti(
                    self._h5Dataset, nodeName=fieldName, subArray=self._h5Dataset[fieldName],
                    fileName=self.fileName, iconColor=self.iconColor))
        return childItems


class H5pyGroupRti(BaseRti):
    """ Repository Tree Item (RTI) that contains a HDF-5 group.
    """
    _defaultIconGlyph = RtiIconFactory.FOLDER

    def __init__(self, h5Group, nodeName, fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor
        """
        super(H5pyGroupRti, self).__init__(nodeName, fileName=fileName, iconColor=iconColor)
        check_class(h5Group, h5py.Group, allow_none=True)

        self._h5Group = h5Group


    @property
    def attributes(self):
        """ The attributes dictionary.
        """
        return attrsToDict(self._h5Group.attrs if self._h5Group else {})


    def _fetchAllChildren(self):
        """ Fetches all sub groups and variables that this group contains.
        """
        assert self._h5Group is not None, "dataset undefined (file not opened?)"
        assert self.canFetchChildren(), "canFetchChildren must be True"

        childItems = []

        for childName, h5Child in self._h5Group.items():
            if isinstance(h5Child, h5py.Group):
                childItems.append(H5pyGroupRti(
                    h5Child, nodeName=childName,
                    fileName=self.fileName, iconColor=self.iconColor))
            elif isinstance(h5Child, h5py.Dataset):
                # The shape can be None in case of Null datasets.
                if h5Child.shape is None or len(h5Child.shape) == 0:
                    childItems.append(H5pyScalarRti(
                        h5Child, nodeName=childName,
                        fileName=self.fileName, iconColor=self.iconColor))
                else:
                    childItems.append(H5pyDatasetRti(
                        h5Child, nodeName=childName,
                        fileName=self.fileName, iconColor=self.iconColor))

            elif isinstance(h5Child, h5py.Datatype):
                #logger.debug("Ignored DataType item: {}".format(childName))
                pass
            else:
                logger.warning("Ignored {}. It has an unexpected HDF-5 type: {}"
                            .format(childName, type(h5Child)))

        return childItems



class H5pyFileRti(H5pyGroupRti):
    """ Reads an HDF-5 file using the h5py package.

        See http://www.h5py.org/
    """
    _defaultIconGlyph = RtiIconFactory.FILE

    def __init__(self, nodeName, fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor
        """
        super(H5pyFileRti, self).__init__(None, nodeName, fileName=fileName, iconColor=iconColor)
        self._checkFileExists()

    def _openResources(self):
        """ Opens the root Dataset.
        """
        logger.info("Opening: {}".format(self._fileName))
        if not os.path.isfile(self._fileName):
            raise OSError("{} does not exist or is not a regular file.".format(self._fileName))
        self._h5Group = h5py.File(self._fileName, 'r')


    def _closeResources(self):
        """ Closes the root Dataset.
        """
        logger.info("Closing: {}".format(self._fileName))
        self._h5Group.close()
        self._h5Group = None



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Configuration Tree Items for use in PyQtGraph-based inspectors
"""
from __future__ import division, print_function

import logging
import warnings

import numpy as np
import pyqtgraph as pg

# Using partial as Python closures are late-binding.
# http://docs.python-guide.org/en/latest/writing/gotchas/#late-binding-closures
from functools import partial
from collections import OrderedDict

from cmlib import ColorSelectionWidget, ColorMap, makeColorBarPixmap
from cmlib import CmMetaData, CatalogMetaData

from argos.config.groupcti import GroupCti
from argos.config.abstractcti import AbstractCti, AbstractCtiEditor
from argos.config.boolcti import BoolCti, BoolGroupCti
from argos.config.choicecti import ChoiceCti
from argos.config.intcti import IntCti
from argos.config.floatcti import SnFloatCti, FloatCti
from argos.config.qtctis import PenCti, ColorCti, createPenStyleCti, createPenWidthCti
from argos.config.untypedcti import UntypedCti
from argos.inspector.pgplugins.colorbar import ArgosColorLegendItem
from argos.inspector.pgplugins.pghistlutitem import HistogramLUTItem
from argos.qt import QtCore, QtGui, QtWidgets, QtSlot
from argos.qt.misc import setWidgetSizePolicy
from argos.repo.colors import CmLibModelSingleton, DEFAULT_COLOR_MAP
from argos.utils.cls import check_class
from argos.utils.masks import nanPercentileOfSubsampledArrayWithMask


from pyqtgraph.graphicsItems.GradientEditorItem import Gradients as GRADIENTS

from pgcolorbar.colorlegend import ColorLegendItem

logger = logging.getLogger(__name__)

X_AXIS = pg.ViewBox.XAxis
Y_AXIS = pg.ViewBox.YAxis
BOTH_AXES = pg.ViewBox.XYAxes
VALID_AXIS_NUMBERS = (X_AXIS, Y_AXIS, BOTH_AXES)
VALID_AXIS_POSITIONS =  ('left', 'right', 'bottom', 'top')
NO_LABEL_STR = '-- none --'


class ViewBoxDebugCti(GroupCti):
    """ Read-only config tree for inspecting a PyQtGraph ViewBox
    """
    def __init__(self, nodeName, viewBox, expanded=False):

        super(ViewBoxDebugCti, self).__init__(nodeName, expanded=expanded)

        self.viewBox = viewBox

        self.insertChild(UntypedCti("targetRange", [[0,1], [0,1]],
            doc="Child coord. range visible [[xmin, xmax], [ymin, ymax]]"))

        self.insertChild(UntypedCti("viewRange", [[0,1], [0,1]],
            doc="Actual range viewed"))

        self.insertChild(UntypedCti("xInverted", None))
        self.insertChild(UntypedCti("yInverted", None))
        self.insertChild(UntypedCti("aspectLocked", False,
            doc="False if aspect is unlocked, otherwise float specifies the locked ratio."))
        self.insertChild(UntypedCti("autoRange", [True, True],
            doc="False if auto range is disabled, otherwise float gives the fraction of data that is visible"))
        self.insertChild(UntypedCti("autoPan", [False, False],
            doc="Whether to only pan (do not change scaling) when auto-range is enabled"))
        self.insertChild(UntypedCti("autoVisibleOnly", [False, False],
            doc="Whether to auto-range only to the visible portion of a plot"))
        self.insertChild(UntypedCti("linkedViews", [None, None],
            doc="may be None, 'viewName', or weakref.ref(view) a name string indicates that the view *should* link to another, but no view with that name exists yet."))
        self.insertChild(UntypedCti("mouseEnabled", [None, None]))
        self.insertChild(UntypedCti("mouseMode", None))
        self.insertChild(UntypedCti("enableMenu", None))
        self.insertChild(UntypedCti("wheelScaleFactor", None))
        self.insertChild(UntypedCti("background", None))

        self.limitsItem = self.insertChild(GroupCti("limits"))
        self.limitsItem.insertChild(UntypedCti("xLimits", [None, None],
            doc="Maximum and minimum visible X values "))
        self.limitsItem.insertChild(UntypedCti("yLimits", [None, None],
            doc="Maximum and minimum visible Y values"))
        self.limitsItem.insertChild(UntypedCti("xRange", [None, None],
            doc="Maximum and minimum X range"))
        self.limitsItem.insertChild(UntypedCti("yRange", [None, None],
            doc="Maximum and minimum Y range"))


    def _refreshNodeFromTarget(self):
        """ Updates the config settings
        """
        for key, value in self.viewBox.state.items():
            if key != "limits":
                childItem = self.childByNodeName(key)
                childItem.data = value
            else:
                # limits contains a dictionary as well
                for limitKey, limitValue in value.items():
                    limitChildItem = self.limitsItem.childByNodeName(limitKey)
                    limitChildItem.data = limitValue



def viewBoxAxisRange(viewBox, axisNumber):
    """ Calculates the range of an axis of a viewBox.
    """
    rect = viewBox.childrenBoundingRect() # taken from viewBox.autoRange()
    if rect is not None:
        if axisNumber == X_AXIS:
            return rect.left(), rect.right()
        elif axisNumber == Y_AXIS:
            return rect.bottom(), rect.top()
        else:
            raise ValueError("axisNumber should be 0 or 1, got: {}".format(axisNumber))
    else:
        # Does this happen? Probably when the plot is empty.
        raise AssertionError("No children bbox. Plot range not updated.")


def inspectorDataRange(inspector, percentage, subsample):
    """ Calculates the range from the inspectors' sliced array. Discards percentage of the minimum
        and percentage of the maximum values of the inspector.slicedArray

        Meant to be used with functools.partial for filling the autorange methods combobox.
        The first parameter is an inspector, it's not an array, because we would then have to
        regenerate the range function every time sliced array of an inspector changes.
    """
    logger.debug("Discarding {}% from id: 0x{:08x}".format(percentage, id(inspector.slicedArray)))

    return nanPercentileOfSubsampledArrayWithMask(
        inspector.slicedArray, (percentage, 100-percentage), subsample)


def defaultAutoRangeMethods(inspector, intialItems=None):
    """ Creates an ordered dict with default autorange methods for an inspector.

        :param inspector: the range methods will work on (the sliced array) of this inspector.
        :param intialItems: will be passed on to the  OrderedDict constructor.
    """
    rangeFunctions = OrderedDict({} if intialItems is None else intialItems)
    rangeFunctions['use all data'] = partial(inspectorDataRange, inspector, 0.0)
    for percentage in [0.1, 0.2, 0.5, 1, 2, 5, 10, 20]:
        label = "discard {}%".format(percentage)
        rangeFunctions[label] = partial(inspectorDataRange, inspector, percentage)
    return rangeFunctions


def setXYAxesAutoRangeOn(commonCti, xAxisRangeCti, yAxisRangeCti, axisNumber):
    """ Turns on the auto range of an X and Y axis simultaneously.
        It sets the autoRangeCti.data of the xAxisRangeCti and yAxisRangeCti to True.
        After that, it emits the sigItemChanged signal of the commonCti.

        Can be used with functools.partial to make a slot that atomically resets the X and Y axis.
        That is, only one sigItemChanged will be emitted.

        This function is necessary because, if one would call PgAxisRangeCti.sigItemChanged
        separately on the X and Y axes the sigItemChanged signal would be emitted twice. This in
        not only slower, but autoscaling one axis may slightly change the others range, so the
        second call to sigItemChanged may unset the autorange of the first.

        axisNumber must be one of: 0 (X-axis), 1 (Y-axis), 2, (Both X and Y axes).
    """
    assert axisNumber in VALID_AXIS_NUMBERS, \
        "Axis number should be one of {}, got {}".format(VALID_AXIS_NUMBERS, axisNumber)

    logger.debug("setXYAxesAutoRangeOn, axisNumber: {}".format(axisNumber))
    if axisNumber == X_AXIS or axisNumber == BOTH_AXES:
        xAxisRangeCti.autoRangeCti.data = True

    if axisNumber == Y_AXIS or axisNumber == BOTH_AXES:
        yAxisRangeCti.autoRangeCti.data = True

    commonCti.model.sigItemChanged.emit(commonCti)



class AbstractRangeCti(GroupCti):
    """ Configuration tree item is linked to a target range.

        Is an abstract class. Descendants must override getTargetRange and setTargetRange
    """
    def __init__(self, autoRangeFunctions=None, nodeName='range',
                 expanded=False, paddingDefault=-1):
        """ Constructor.
            The target axis is specified by viewBox and axisNumber (0 for x-axis, 1 for y-axis)

            If given, autoRangeMethods must be a (label to function) dictionary that will be used
            to populate the (auto range) method ChoiceCti.
            If autoRangeFunctions is None, there will be no auto-range child CTI.
            If autoRangeFunctions has one element there will be an auto-range child without a method
            child CTI (the function from the autoRangeMethods dictionary will be the default).

            :param int paddingDefault: default padding value. Use -1 for 'dynamic' padding, which
                is the PyQtGraph padding algorithm.
        """
        super(AbstractRangeCti, self).__init__(nodeName, expanded=expanded)

        self._rangeFunctions = {}
        self.autoRangeCti = None
        self.methodCti = None
        self.subsampleCti = None
        self.paddingCti = None

        if autoRangeFunctions is not None:
            self.autoRangeCti = self.insertChild(BoolCti("auto-range", True, expanded=False))
            self._rangeFunctions = autoRangeFunctions

            if len(autoRangeFunctions) > 1:
                self.methodCti = ChoiceCti("method", configValues=list(autoRangeFunctions.keys()))
                self.autoRangeCti.insertChild(self.methodCti)

                self.subsampleCti = BoolCti("subsample", True)
                self.autoRangeCti.insertChild(self.subsampleCti)

            self.paddingCti = IntCti("padding", paddingDefault,
                                     suffix="%", specialValueText="dynamic",
                                     minValue=-1, maxValue=1000, stepSize=1)
            self.autoRangeCti.insertChild(self.paddingCti)

        self.rangeMinCti = self.insertChild(SnFloatCti('min', 0.0))
        self.rangeMaxCti = self.insertChild(SnFloatCti('max', 1.0))


    @property
    def autoRangeMethod(self):
        """ The currently selected auto range method.
            If there is no method child CTI, there will be only one method (which will be returned).
        """
        if self.methodCti:
            return self.methodCti.configValue
        else:
            assert len(self._rangeFunctions) == 1, \
                "Assumed only one _rangeFunctions. Got: {}".format(self._rangeFunctions)
            return list(self._rangeFunctions.keys())[0]


    def _refreshNodeFromTarget(self, *args, **kwargs):
        """ Used to update the axis config tree item when the target axes was changed.
            It updates the autoRange checkbox (_forceRefreshAutoRange) and calculates new min max config
            values from range by calling _forceRefreshMinMax.

            The *args and **kwargs arguments are ignored.
        """
        # No need to check for self.getRefreshBlocked().
        # This is already done in the caller (refreshFromTarget)
        self._forceRefreshAutoRange()
        self._forceRefreshMinMax()


    def _forceRefreshMinMax(self):
        """ Refreshes the min max config values from the axes' state.
        """
        #logger.debug("_forceRefreshMinMax")

        # Set the precision from by looking how many decimals are needed to show the difference
        # between the minimum and maximum, given the maximum. E.g. if min = 0.04 and max = 0.07,
        # we would only need zero decimals behind the point as we can write the range as
        # [4e-2, 7e-2]. However if min = 1.04 and max = 1.07, we need 2 decimals behind the point.
        # So, while the range is the same size we need more decimals because we are not zoomed in
        # around zero.
        rangeMin, rangeMax = self.getTargetRange() # [[xmin, xmax], [ymin, ymax]]
        with warnings.catch_warnings():
            # Ignore divide by zero warnings when all elements have the same value
            warnings.simplefilter("ignore")
            maxOrder = np.log10(np.abs(max(rangeMax, rangeMin)))
            diffOrder = np.log10(np.abs(rangeMax - rangeMin))

            extraDigits = 2 # add some extra digits to make each pan/zoom action show a new value.
            precisionF = np.clip(abs(maxOrder - diffOrder) + extraDigits, extraDigits + 1, 25)
            precision = int(precisionF) if np.isfinite(precisionF) else extraDigits + 1
            #logger.debug("maxOrder: {}, diffOrder: {}, precision: {}"
            #             .format(maxOrder, diffOrder, precision))

        self.rangeMinCti.precision = precision
        self.rangeMaxCti.precision = precision
        self.rangeMinCti.data, self.rangeMaxCti.data = rangeMin, rangeMax

        # Update values in the tree
        self.model.emitDataChanged(self.rangeMinCti)
        self.model.emitDataChanged(self.rangeMaxCti)


    def refreshMinMax(self):
        """ Refreshes the min max config values from the axes' state.
            Does nothing when self.getRefreshBlocked() returns True.
        """
        if self.getRefreshBlocked():
            #logger.debug("refreshMinMax blocked for {}".format(self.nodeName))
            return

        self._forceRefreshMinMax()


    def _forceRefreshAutoRange(self):
        """ The min and max config items will be disabled if auto range is on.
        """
        enabled = self.autoRangeCti and self.autoRangeCti.configValue
        self.rangeMinCti.enabled = not enabled
        self.rangeMaxCti.enabled = not enabled
        self.model.emitDataChanged(self)


    def setAutoRangeOff(self):
        """ Turns off the auto range checkbox.
            Calls _refreshNodeFromTarget, not _updateTargetFromNode, because setting auto range off
            does not require a redraw of the target.
        """
        # TODO: catch exceptions. How?
        # /argos/hdf-eos/DeepBlue-SeaWiFS-1.0_L3_20100101_v002-20110527T191319Z.h5/aerosol_optical_thickness_stddev_ocean
        if self.getRefreshBlocked():
            logger.debug("setAutoRangeOff blocked for {}".format(self.nodeName))
            return

        if self.autoRangeCti:
            self.autoRangeCti.data = False

        self._forceRefreshAutoRange()


    def setAutoRangeOn(self):
        """ Turns on the auto range checkbox for the equivalent axes
            Emits the sigItemChanged signal so that the inspector may be updated.

            Use the setXYAxesAutoRangeOn stand-alone function if you want to set the autorange on
            for both axes of a viewport.
        """
        if self.getRefreshBlocked():
            logger.debug("Set autorange on blocked for {}".format(self.nodeName))
            return

        if self.autoRangeCti:
            self.autoRangeCti.data = True
        self.model.sigItemChanged.emit(self) # this should typically only be called by other classes.


    def calculateRange(self):
        """ Calculates the range depending on the config settings.
        """
        if not self.autoRangeCti or not self.autoRangeCti.configValue:
            return (self.rangeMinCti.data, self.rangeMaxCti.data)
        else:
            rangeFunction = self._rangeFunctions[self.autoRangeMethod]

            if self.subsampleCti is None:
                return rangeFunction()
            else:
                return rangeFunction(self.subsampleCti.configValue)


    def _updateTargetFromNode(self):
        """ Applies the configuration to the target axis.
        """
        if not self.autoRangeCti or not self.autoRangeCti.configValue:
            padding = 0
        elif self.paddingCti.configValue == -1: # specialValueText
            # PyQtGraph dynamic padding: between 0.02 and 0.1 dep. on the size of the ViewBox
            padding = None
        else:
            padding = self.paddingCti.configValue / 100

        targetRange = self.calculateRange()
        #logger.debug("axisRange: {}, padding={}".format(targetRange, padding))
        if not np.all(np.isfinite(targetRange)):
            logger.debug("New target range is not finite. Plot range not updated")
            return

        self.setTargetRange(targetRange, padding=padding)


    def getTargetRange(self):
        """ Gets the range of the target
        """
        raise NotImplementedError("Abstract method. Please override.")


    def setTargetRange(self, targetRange, padding=None):
        """ Sets the range of the target.
        """
        # The padding parameter is a bit of a hack.
        # TODO: move to PgAxisRangeCti or implement in PgHistLutColorRangeCti?
        # That last option may be useful to colorize images with uints, which currently don't
        # show out black pixels for the maximum values (e.g. 255)
        raise NotImplementedError("Abstract method. Please override.")



class PgAxisRangeCti(AbstractRangeCti):
    """ Configuration tree item is linked to the axis range.
    """
    PYQT_RANGE = 'by PyQtGraph'

    def __init__(self, viewBox, axisNumber, autoRangeFunctions=None,
                 nodeName='range', expanded=True):
        """ Constructor.
            The target axis is specified by viewBox and axisNumber (0 for x-axis, 1 for y-axis)

            If given, autoRangeFunctions must be a (label to function) dictionary that will be used
            to populate the (auto range) method ChoiceCti. If not give, the there will not be
            a method choice and the autorange implemented by PyQtGraph will be used.
        """
        if autoRangeFunctions is None:
            autoRangeFunctions = {self.PYQT_RANGE: partial(viewBoxAxisRange, viewBox, axisNumber)}

        super(PgAxisRangeCti, self).__init__(autoRangeFunctions=autoRangeFunctions,
                                             nodeName=nodeName, expanded=expanded)
        check_class(viewBox, pg.ViewBox)
        assert axisNumber in (X_AXIS, Y_AXIS), "axisNumber must be 0 or 1"

        self.viewBox = viewBox
        self.axisNumber = axisNumber

        # Autorange must be disabled as not to interfere with this class.
        # Note that autorange of ArgosPgPlotItem is set to False by default.
        axisAutoRange = self.viewBox.autoRangeEnabled()[axisNumber]
        assert axisAutoRange is False, \
            "Autorange is {!r} for axis {} of {}".format(axisAutoRange, axisNumber, self.nodePath)

        # Connect signals
        self.viewBox.sigRangeChangedManually.connect(self.setAutoRangeOff)
        self.viewBox.sigRangeChanged.connect(self.refreshMinMax)


    def _closeResources(self):
        """ Disconnects signals.
            Is called by self.finalize when the cti is deleted.
        """
        self.viewBox.sigRangeChangedManually.disconnect(self.setAutoRangeOff)
        self.viewBox.sigRangeChanged.disconnect(self.refreshMinMax)


    def getTargetRange(self):
        """ Gets the range of the target
        """
        return self.viewBox.state['viewRange'][self.axisNumber]


    def setTargetRange(self, targetRange, padding=None):
        """ Sets the range of the target.
        """
        # viewBox.setRange doesn't accept an axis number :-(
        if self.axisNumber == X_AXIS:
            xRange, yRange = targetRange, None
        else:
            xRange, yRange = None, targetRange

        # Do not set disableAutoRange to True in setRange; it triggers 'one last' auto range.
        # This is why the viewBox' autorange must be False at construction.
        self.viewBox.setRange(xRange = xRange, yRange=yRange, padding=padding,
                              update=False, disableAutoRange=False)

        # Sanity check (can only work if we calculate padding ourselves)
        #if self.viewBox.state['viewRange'][self.axisNumber] != targetRange:
        #    raise AssertionError("ViewRange update failed: {} != {}"
        #        .format(self.viewBox.state['viewRange'][self.axisNumber], targetRange))



class PgHistLutColorRangeCti(AbstractRangeCti):
    """ Configuration tree item is linked to the HistogramLUTItem range.

        Used in the old imageplot2d.py
    """
    def __init__(self, histLutItem, autoRangeFunctions=None, nodeName='color range', expanded=True):
        """ Constructor.
            The target axis is specified by viewBox and axisNumber (0 for x-axis, 1 for y-axis)

            If given, autoRangeFunctions must be a (label to function) dictionary that will be used
            to populate the (auto range) method ChoiceCti. If not give, the there will not be
            a method choice and the autorange implemented by PyQtGraph will be used.
        """
        super(PgHistLutColorRangeCti, self).__init__(autoRangeFunctions=autoRangeFunctions,
                                                     nodeName=nodeName, expanded=expanded)
        check_class(histLutItem, HistogramLUTItem)
        self.histLutItem = histLutItem

        # Connect signals
        self.histLutItem.sigLevelsChanged.connect(self.setAutoRangeOff)
        self.histLutItem.sigLevelsChanged.connect(self.refreshMinMax)


    def _closeResources(self):
        """ Disconnects signals.
            Is called by self.finalize when the cti is deleted.
        """
        self.histLutItem.sigLevelsChanged.disconnect(self.setAutoRangeOff)
        self.histLutItem.sigLevelsChanged.disconnect(self.refreshMinMax)


    def getTargetRange(self):
        """ Gets the (color) range of the HistogramLUTItem
        """
        return self.histLutItem.getLevels()


    def setTargetRange(self, targetRange, padding=None):
        """ Sets the (color) range of the HistogramLUTItem
            The padding variable is ignored.
        """
        rangeMin, rangeMax = targetRange
        self.histLutItem.setLevels(rangeMin, rangeMax)



class PgColorLegendCti(AbstractRangeCti):
    """ Configuration tree item is linked to the HistogramLUTItem range.
    """
    def __init__(self, legend, autoRangeFunctions=None, nodeName='color range', expanded=True):
        """ Constructor.
            The target axis is specified by viewBox and axisNumber (0 for x-axis, 1 for y-axis)

            If given, autoRangeFunctions must be a (label to function) dictionary that will be used
            to populate the (auto range) method ChoiceCti. If not give, the there will not be
            a method choice and the autorange implemented by PyQtGraph will be used.
        """
        super(PgColorLegendCti, self).__init__(
            autoRangeFunctions=autoRangeFunctions, nodeName=nodeName,
            expanded=True, paddingDefault=0)

        check_class(legend, ColorLegendItem)
        self.legend = legend

        self.paddingCti.defaultData = 0
        # Connect signals
        self.legend.sigLevelsChanged.connect(self.setAutoRangeOff)
        self.legend.sigLevelsChanged.connect(self.refreshMinMax)


    def _closeResources(self):
        """ Disconnects signals.
            Is called by self.finalize when the cti is deleted.
        """
        self.legend.sigLevelsChanged.disconnect(self.setAutoRangeOff)
        self.legend.sigLevelsChanged.disconnect(self.refreshMinMax)


    def getTargetRange(self):
        """ Gets the (color) range of the HistogramLUTItem
        """
        return self.legend.getLevels()


    def setTargetRange(self, targetRange, padding=None):
        """ Sets the (color) range of the HistogramLUTItem
        """
        self.legend.setLevels(targetRange, padding=padding)



class PgColorLegendLabelCti(ChoiceCti):
    """ Configuration tree item that is linked to the axis label of a color legend .
    """
    def __init__(self, colorLegendItem, collector,
                 nodeName='label', defaultData=0, configValues=None):
        """ Constructor
            :param colorLegendItem PgColorLegendLabelCti:
            :param collector: needed to get the collector.rtiInfo
            :param nodeName: the node name of this config tree item (default = label
            :param defaultData:
            :param configValues:
        """
        super(PgColorLegendLabelCti, self).__init__(
            nodeName, editable=True, defaultData=defaultData, configValues=configValues)

        check_class(colorLegendItem, ColorLegendItem)
        self.colorLegendItem = colorLegendItem
        self.collector = collector


    def _updateTargetFromNode(self):
        """ Applies the configuration to the target axis it monitors.
            The axis label will be set to the configValue. If the configValue equals
            NO_LABEL_STR, the label will be hidden.
        """
        if self.configValue == NO_LABEL_STR:
            self.colorLegendItem.setLabel(None)
        else:
            rtiInfo = self.collector.rtiInfo
            self.colorLegendItem.setLabel(self.configValue.format(**rtiInfo))



class PgShowHistCti(BoolCti):
    """ BoolCti that shows/hides the histogram of a color bar.
    """
    def __init__(self, colorLegendItem, nodeName='show histogram', defaultData=True):
        """ Constructor.
            The target axis is specified by viewBox and axisNumber (0 for x-axis, 1 for y-axis)
        """
        super(PgShowHistCti, self).__init__(nodeName, defaultData=defaultData)

        check_class(colorLegendItem, ColorLegendItem)
        self.colorLegendItem = colorLegendItem


    def _updateTargetFromNode(self):
        """ Applies the configuration to its target axis
        """
        self.colorLegendItem.showHistogram(self.configValue)


class PgShowDragLinesCti(BoolCti):
    """ BoolCti that shows/hides the drag lines a color bar.
    """
    def __init__(self, colorLegendItem, nodeName='show drag lines',
                 defaultData=True, expanded=False):
        """ Constructor.
            The target axis is specified by viewBox and axisNumber (0 for x-axis, 1 for y-axis)
        """
        super(PgShowDragLinesCti, self).__init__(nodeName,
                                                 defaultData=defaultData, expanded=expanded)

        self.marginCti = self.insertChild(
            IntCti("margins", 40, minValue=0, maxValue=250, stepSize=5))

        check_class(colorLegendItem, ColorLegendItem)
        self.colorLegendItem = colorLegendItem


    def _updateTargetFromNode(self):
        """ Applies the configuration to its target axis
        """
        try:
            self.colorLegendItem.showDragLines(self.configValue)
            self.colorLegendItem.setEdgeMargins(self.marginCti.configValue)
        except AttributeError as ex:
            logger.warning("Please update cmlib to latest version. No drag lines: {}".format(ex))


class PgGradientEditorItemCti(ChoiceCti):
    """ Lets the user select one of the standard color scales in a GradientEditorItem

        Is not used in the new PgImagePlot2d, only in the one in old_imageplot2d.py.
    """
    def __init__(self, gradientEditorItem, nodeName="color scale", defaultData=-1):
        """ Constructor.
            The gradientEditorItem must be a PyQtGraph.GradientEditorItem.
            The configValues are taken from pyqtgraph.graphicsItems.GradientEditorItem.Gradients
            which is an OrderedDict of color scales. By default the last item from this list is
            chosen, which is they 'grey' color scale.
        """
        super(PgGradientEditorItemCti, self).__init__(nodeName, defaultData=defaultData,
                                                      configValues=list(GRADIENTS.keys()))
        check_class(gradientEditorItem, pg.GradientEditorItem)
        self.gradientEditorItem = gradientEditorItem


    def _updateTargetFromNode(self):
        """ Applies the configuration to its target widget
        """
        self.gradientEditorItem.loadPreset(self.configValue)



class PgAspectRatioCti(BoolCti):
    """ BoolCti for locking and specifying the aspect ratio (x/y)
    """
    def __init__(self, viewBox, nodeName="lock aspect ratio", defaultData=False, expanded=False):
        """ Constructor.
            The target axis is specified by viewBox and axisNumber (0 for x-axis, 1 for y-axis)
        """
        super(PgAspectRatioCti, self).__init__(nodeName, defaultData=defaultData, expanded=expanded)
        check_class(viewBox, pg.ViewBox)

        self.aspectRatioCti = self.insertChild(FloatCti("ratio", 1.0, minValue=0.0))
        self.viewBox = viewBox


    def _updateTargetFromNode(self):
        """ Applies the configuration to its target axis
        """
        self.viewBox.setAspectLocked(lock=self.configValue, ratio=self.aspectRatioCti.configValue)



class PgAxisFlipCti(BoolCti):
    """ BoolCti that flips an axis when True.
    """
    def __init__(self, viewBox, axisNumber, nodeName='flipped', defaultData=False
                 ):
        """ Constructor.
            The target axis is specified by viewBox and axisNumber (0 for x-axis, 1 for y-axis)
        """
        super(PgAxisFlipCti, self).__init__(nodeName, defaultData=defaultData)
        check_class(viewBox, pg.ViewBox)
        assert axisNumber in (X_AXIS, Y_AXIS), "axisNumber must be 0 or 1"
        self.viewBox = viewBox
        self.axisNumber = axisNumber


    def _updateTargetFromNode(self):
        """ Applies the configuration to its target axis
        """
        if self.axisNumber == X_AXIS:
            self.viewBox.invertX(self.configValue)
        else:
            self.viewBox.invertY(self.configValue)



class PgAxisLabelCti(ChoiceCti):
    """ Configuration tree item that is linked to the axis label.
    """
    def __init__(self, plotItem, axisPosition, collector,
                 nodeName='label', defaultData=0, configValues=None):
        """ Constructor
            :param plotItem:
            :param axisPosition: 'left', 'right', 'bottom', 'top'
            :param collector: needed to get the collector.rtiInfo
            :param nodeName: the node name of this config tree item (default = label
            :param defaultData:
            :param configValues:
        """
        super(PgAxisLabelCti, self).__init__(nodeName, editable=True,
                                             defaultData=defaultData, configValues=configValues)
        assert axisPosition in VALID_AXIS_POSITIONS,\
            "axisPosition must be in {}".format(VALID_AXIS_POSITIONS)
        self.collector = collector
        self.plotItem = plotItem
        self.axisPosition = axisPosition


    def _updateTargetFromNode(self):
        """ Applies the configuration to the target axis it monitors.
            The axis label will be set to the configValue. If the configValue equals
             NO_LABEL_STR, the label will be hidden.
        """
        rtiInfo = self.collector.rtiInfo
        self.plotItem.setLabel(self.axisPosition, self.configValue.format(**rtiInfo))
        self.plotItem.showLabel(self.axisPosition, self.configValue != NO_LABEL_STR)



class PgAxisShowCti(BoolCti):
    """ BoolCti that toggles axisPosition showing/hiding an axis.
    """
    def __init__(self, plotItem, axisPosition, nodeName='show', defaultData=True):
        """ Constructor.
            The target axis is specified by imagePlotItem and axisPosition.
            axisPosition must be one of: 'left', 'right', 'bottom', 'top'

            NOTE: the PyQtGraph showAxis seems not to work. TODO: investigate.
        """
        super(PgAxisShowCti, self).__init__(nodeName, defaultData=defaultData)
        assert axisPosition in VALID_AXIS_POSITIONS,\
            "axisPosition must be in {}".format(VALID_AXIS_POSITIONS)

        self.plotItem = plotItem
        self.axisPosition = axisPosition


    def _updateTargetFromNode(self):
        """ Applies the configuration to its target axis
        """
        logger.debug("showAxis: {}, {}".format(self.axisPosition, self.configValue))
        self.plotItem.showAxis(self.axisPosition, show=self.configValue) # Seems not to work



class PgAxisLogModeCti(BoolCti):
    """ BoolCti that toggles an axis between logarithmic vs linear mode.
    """
    def __init__(self, plotItem, axisNumber, nodeName='logarithmic', defaultData=False):
        """ Constructor.
            The target axis is specified by imagePlotItem and axisNumber (0 for x-axis, 1 for y-axis)
        """
        super(PgAxisLogModeCti, self).__init__(nodeName, defaultData=defaultData)
        self.plotItem = plotItem
        self.axisNumber = axisNumber


    def _updateTargetFromNode(self):
        """ Applies the configuration to its target axis
        """
        if self.axisNumber == X_AXIS:
            xMode, yMode = self.configValue, None
        else:
            xMode, yMode = None, self.configValue

        self.plotItem.setLogMode(x=xMode, y=yMode)



class PgGridCti(BoolGroupCti):
    """ CTI for toggling the the grid on and off.

        Has child CTIs for toggling the X and Y axes separately. These are typically not added
        as children of PgAxisCti objects so that the user can enable both grids with one checkbox.
        Also includes a child to configure the transparency (alpha) of the grid.
    """
    def __init__(self, plotItem, nodeName="grid", defaultData=True, expanded=False):
        """ Constructor.
            The target axis is specified by viewBox and axisNumber (0 for x-axis, 1 for y-axis)
        """
        super(PgGridCti, self).__init__(nodeName, defaultData=defaultData, expanded=expanded)
        check_class(plotItem, pg.PlotItem)
        self.plotItem = plotItem

        self.xGridCti = self.insertChild(BoolCti('x-axis', defaultData))
        self.yGridCti = self.insertChild(BoolCti('y-axis', defaultData))
        self.alphaCti = self.insertChild(FloatCti('alpha', 0.20,
            minValue=0.0, maxValue=1.0, stepSize=0.01, decimals=2))


    def _updateTargetFromNode(self):
        """ Applies the configuration to the grid of the plot item.
        """
        self.plotItem.showGrid(x=self.xGridCti.configValue, y=self.yGridCti.configValue,
                               alpha=self.alphaCti.configValue)
        self.plotItem.updateGrid()



class PgAxisCti(GroupCti):
    """ Configuration tree item for a PyQtGraph plot axis.

        Currently nothing more than a GroupCti.
    """
    pass



class PgPlotDataItemCti(GroupCti):
    """ Configuration tree item for a PyQtGraph plot data item.
        It allows for configuring a line style, color and symbols.

        Note that using a line width > 1 in comination with anti-aliassing may be slow.
        See https://github.com/pyqtgraph/pyqtgraph/issues/533
    """
    def __init__(self, nodeName="pen",  defaultData=None, expanded=True):
        """ Constructor.
        """
        super(PgPlotDataItemCti, self).__init__(nodeName, defaultData=defaultData, expanded=expanded)
        self.antiAliasCti = self.insertChild(BoolCti("anti-alias", True))

        self.colorCti = self.insertChild(ColorCti('color', QtGui.QColor('#FF0066')))
        self.lineCti = self.insertChild(BoolCti('line', True, expanded=False,
                                                childrenDisabledValue=False))
        self.lineStyleCti = self.lineCti.insertChild(createPenStyleCti('style'))
        self.lineWidthCti = self.lineCti.insertChild(createPenWidthCti('width'))

        defaultShadowPen = QtGui.QPen(QtGui.QColor('#BFBFBF'))
        defaultShadowPen.setWidth(0)
        self.lineCti.insertChild(PenCti("shadow", False, expanded=False,
                                        resetTo=QtGui.QPen(defaultShadowPen),
                                        includeNoneStyle=True, includeZeroWidth=True))

        self.symbolCti = self.insertChild(BoolCti("symbol", False, expanded=False,
                                             childrenDisabledValue=False))
        self.symbolShapeCti = self.symbolCti.insertChild(ChoiceCti("shape", 0,
           displayValues=['circle', 'square', 'triangle', 'diamond', 'plus'],
           configValues=['o', 's', 't', 'd', '+']))
        self.symbolSizeCti = self.symbolCti.insertChild(IntCti('size', 5, minValue=0,
                                                               maxValue=100, stepSize=1))

    @property
    def penColor(self):
        """ Returns the pen/color value"""
        return self.colorCti.configValue


    def createPlotDataItem(self):
        """ Creates a PyQtGraph PlotDataItem from the config values
        """
        antialias = self.antiAliasCti.configValue

        color = self.penColor
        if self.lineCti.configValue:
            pen = QtGui.QPen()
            pen.setCosmetic(True)
            pen.setColor(color)
            pen.setWidthF(self.lineWidthCti.configValue)
            pen.setStyle(self.lineStyleCti.configValue)
            shadowCti = self.lineCti.findByNodePath('shadow')
            shadowPen = shadowCti.createPen(altStyle=pen.style(), altWidth=2.0 * pen.widthF())
        else:
            pen = None
            shadowPen = None

        drawSymbols = self.symbolCti.configValue
        symbolShape = self.symbolShapeCti.configValue if drawSymbols else None
        symbolSize  = self.symbolSizeCti.configValue if drawSymbols else 0.0
        symbolPen = None # otherwise the symbols will also have dotted/solid line.
        symbolBrush = QtGui.QBrush(color) if drawSymbols else None

        plotDataItem = pg.PlotDataItem(antialias=antialias, pen=pen, shadowPen=shadowPen,
                                       symbol=symbolShape, symbolSize=symbolSize,
                                       symbolPen=symbolPen, symbolBrush=symbolBrush)
        return plotDataItem



class PgColorMapCti(AbstractCti):
    """ Lets the user select one of color maps of the color map library
    """
    SUB_SAMPLING_OFF = 1

    def __init__(self, colorLegendItem,
                 nodeName="color map", defaultData=DEFAULT_COLOR_MAP, expanded=False):
        """ Constructor.

            Stores a color map as data.

            :param defaultData: the default index in the combobox that is used for editing
        """
        check_class(colorLegendItem, ArgosColorLegendItem)
        self.colorLegendItem = colorLegendItem
        self.cmLibModel = CmLibModelSingleton.instance()

        # grey scale color map for when no color map is selected.
        lutRgba = np.outer(np.arange(256, dtype=np.uint8), np.array([1, 1, 1, 1], dtype=np.uint8))
        lutRgba[:,3] = 255
        self.greyScaleColorMap = ColorMap(CmMetaData("-- none --"), CatalogMetaData("Argos"))
        self.greyScaleColorMap.set_rgba_uint8_array(lutRgba)

        super(PgColorMapCti, self).__init__(nodeName, defaultData, expanded=expanded)

        self.reverseCti = self.insertChild(
            BoolCti("reversed", False))

        self.subSampleCti = self.insertChild(
            IntCti("subsample", self.SUB_SAMPLING_OFF, specialValueText="off",
                   minValue=self.SUB_SAMPLING_OFF, maxValue=64, stepSize=1)
        )


    def _enforceDataType(self, data):
        """ Converts to int so that this CTI always stores that type.
        """
        if data is None:
            return self.greyScaleColorMap
        elif isinstance(data, ColorMap):
            return data
        else:
            return self.cmLibModel.getColorMapByKey(data)


    @property
    def configValue(self):
        """ The currently selected configValue

            :rtype: ColorMap
        """
        return self.data


    def _updateTargetFromNode(self):
        """ Applies the configuration to its target axis.
            Sets the image item's lookup table to the LUT of the selected color map.
        """
        lut = self.data.rgb_uint8_array

        targetSize = self.subSampleCti.configValue
        if targetSize > self.SUB_SAMPLING_OFF:
            sourceSize, _ = lut.shape
            subIdx = np.round(np.linspace(0, sourceSize-1, targetSize)).astype(np.uint)
            lut = lut[subIdx, :]

        if self.reverseCti.configValue:
            lut = np.flipud(lut)

        self.colorLegendItem.setLut(lut)


    def _dataToString(self, data):
        """ Conversion function used to convert the (default)data to the display value.
        """
        return "" if data is None else data.meta_data.pretty_name


    @property
    def decoration(self):
        """ Returns a pixmap of the color map to show as icon
        """
        return makeColorBarPixmap(
            self.data,
            width=self.cmLibModel.iconBarWidth * 0.65,
            height=self.cmLibModel.iconBarHeight * 0.65,
            drawBorder=self.cmLibModel.drawIconBarBorder)


    def _nodeMarshall(self):
        """ Returns the non-recursive marshalled value of this CTI. Is called by marshall()
        """
        return self.data.key


    def _nodeUnmarshall(self, key):
        """ Initializes itself non-recursively from data. Is called by unmarshall()
        """
        self.data = self.cmLibModel.getColorMapByKey(key)


    def createEditor(self, delegate, parent, option):
        """ Creates a ChoiceCtiEditor.
            For the parameters see the AbstractCti constructor documentation.
        """
        return PgColorMapCtiEditor(self, delegate, parent=parent)



class PgColorMapCtiEditor(AbstractCtiEditor):
    """ A CtiEditor which contains a QCombobox for editing ChoiceCti objects.
    """
    def __init__(self, cti, delegate, parent=None):
        """ See the AbstractCtiEditor for more info on the parameters
        """
        super(PgColorMapCtiEditor, self).__init__(cti, delegate, parent=parent)

        selectionWidget = ColorSelectionWidget(self.cti.cmLibModel)
        setWidgetSizePolicy(selectionWidget, QtWidgets.QSizePolicy.Expanding, None)

        selectionWidget.sigColorMapHighlighted.connect(self.onColorMapHighlighted)
        selectionWidget.sigColorMapChanged.connect(self.onColorMapChanged)
        self.selectionWidget = self.addSubEditor(selectionWidget, isFocusProxy=True)
        self.comboBox = self.selectionWidget.comboBox


    def finalize(self):
        """ Is called when the editor is closed. Disconnect signals.
        """
        logger.debug("PgColorMapCtiEditor.finalize")
        self.selectionWidget.sigColorMapChanged.disconnect(self.onColorMapChanged)
        self.selectionWidget.sigColorMapHighlighted.disconnect(self.onColorMapHighlighted)
        super(PgColorMapCtiEditor, self).finalize()


    def setData(self, data):
        """ Provides the main editor widget with a data to manipulate.
        """
        if data is None:
            logger.warning("No color map to select")
        else:
            self.selectionWidget.setColorMapByKey(data.key)


    def getData(self):
        """ Gets data from the editor widget.
        """
        return self.selectionWidget.getCurrentColorMap()


    @QtSlot(ColorMap)
    def onColorMapHighlighted(self, colorMap):
        """ Is called when the user highlights an item in the combo box or dialog.

            The item's index is passed.
            Note that this signal is sent even when the choice is not changed.
        """
        logger.debug("onColorMapHighlighted({})".format(colorMap))
        self.cti.data = colorMap
        self.cti.updateTarget()


    @QtSlot(ColorMap)
    def onColorMapChanged(self, index):
        """ Is called when the user chooses an item in the combo box. The item's index is passed.
            Note that this signal is sent even when the choice is not changed.
        """
        logger.debug("onColorMapChanged")
        self.delegate.commitData.emit(self)




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Module that contains ArgosPgColorBar.
"""
from __future__ import division, print_function

import logging
import warnings

import pyqtgraph as pg

from argos.qt import QtCore, QtWidgets, QtSignal


from argos.utils.cls import check_class
from pgcolorbar.colorlegend import ColorLegendItem

# from argos.info import DEBUGGING
# from argos.qt import Qt, QtCore, QtWidgets, QtSignal

logger = logging.getLogger(__name__)

warnings.filterwarnings(action='default', category=RuntimeWarning, module='pgcolorbar.colorlegend')  # Show once


class ArgosColorLegendItem(ColorLegendItem):
    """ Wrapper around pgcolorbar.colorlegend.ColorLegendItem.

        Suppresses the FutureWarning of PyQtGraph in _updateHistogram.
        Overrides the _imageItemHasIntegerData method.
        Adds context menu with reset color scale action.
        Middle mouse click resets the axis with the settings in the config tree.
    """
    # Use a QueuedConnection to connect to sigResetColorScale so that the reset is scheduled after
    # all current events have been processed. Otherwise the mouseReleaseEvent may look for a
    # PlotCurveItem that is no longer present after the reset, which results in a RuntimeError:
    # wrapped C/C++ object of type PlotCurveItem has been deleted.
    sigResetColorScale = QtSignal()  # Signal the inspectors to reset the color scale


    def __init__(self, *args, histHeightPercentile=99.0, **kwargs):
        """ Constructor
        """
        super(ArgosColorLegendItem, self).__init__(
            *args, histHeightPercentile=histHeightPercentile, **kwargs)
        self.resetColorScaleAction = QtWidgets.QAction("Reset Color Range", self)
        self.resetColorScaleAction.triggered.connect(self.emitResetColorScaleSignal)
        self.resetColorScaleAction.setToolTip("Reset the range of the color scale.")
        self.addAction(self.resetColorScaleAction)


    @classmethod
    def _imageItemHasIntegerData(cls, imageItem):
        """ Returns True if the imageItem contains integer data.

            Overriden so that the ImagePlotItem can replace integer arrays with float arrays (to
            plot masked values as NaNs) while it still calculates the histogram bins as if it
            where integers (to prevent aliasing)
        """
        check_class(imageItem, pg.ImageItem, allow_none=True)

        if hasattr(imageItem, '_wasIntegerData'):
            return imageItem._wasIntegerData
        else:
            return super(ArgosColorLegendItem, cls)._imageItemHasIntegerData(imageItem)


    def emitResetColorScaleSignal(self):
        """ Emits the sigColorScaleReset to request the inspectors to reset the color scale
        """
        logger.debug("Emitting sigColorScaleReset() for {!r}".format(self))
        self.sigResetColorScale.emit()


    def mouseClickEvent(self, mouseClickEvent):
        """ Handles (PyQtGraph) mouse click events.

            Overrides the middle mouse click to reset using the settings in the config tree.

            Opens the context menu if a right mouse button was clicked. (We can't simply use
            setContextMenuPolicy(Qt.ActionsContextMenu because the PlotItem class does not derive
            from QWidget).

            :param mouseClickEvent: pyqtgraph.GraphicsScene.mouseEvents.MouseClickEvent
        """
        if mouseClickEvent.button() in self.resetRangeMouseButtons:
            self.emitResetColorScaleSignal()
            mouseClickEvent.accept()

        elif mouseClickEvent.button() == QtCore.Qt.RightButton:
            contextMenu = QtWidgets.QMenu()
            for action in self.actions():
                contextMenu.addAction(action)

            screenPos = mouseClickEvent.screenPos() # Screenpos is a QPointF, convert to QPoint.
            screenX = round(screenPos.x())
            screenY = round(screenPos.y())
            contextMenu.exec_(QtCore.QPoint(screenX, screenY))

        else:
            super(ArgosColorLegendItem, self).mouseClickEvent(mouseClickEvent)




# -*- coding: utf-8 -*-

""" Various function related to config files.
"""
import logging
from typing import Dict, Any

logger = logging.getLogger(__name__)

ConfigDict = Dict[str, Any]

def __findConfigParameter(cfg, parts, orgPath):
    """ Recursively finds a parameter in a (config) dict.
    """
    if len(parts) == 0:
        return cfg  # Trailing slash

    head, tail = parts[0], parts[1:]

    if head == "":  # A leading slash or double orgPath
        return __findConfigParameter(cfg, tail, orgPath)

    if head not in cfg:
        msg = "Path {!r} not found in dict. {!r} not in: {}".format(orgPath, head, list(cfg.keys()))
        raise KeyError(msg)

    if len(tail) == 0:
        return cfg[head]
    else:
        return __findConfigParameter(cfg[head], tail, orgPath)


def findConfigParameter(cfg, path):
    """ Recursively finds a parameter in a config dict.

        Raises KeyError if the path cannot be found.

        :param cfg: config dictionary. Can be a recursive dict (a dict of dicts, etc)
        :param path: Slash separated parameter path. E.g.: '/dict1/dict2/parameter'
    """
    if not path:
        # An empty path is most likely a programming error
        raise KeyError("Empty path given to findConfigParameter: {}".format(path))
    else:
        parts = path.split('/')
        return __findConfigParameter(cfg, parts, path)


def getConfigParameter(cfg, path, alt=None):
    """ Finds the findConfigParameter. Returns alternative value if not found.

        Will still log a warning if the parameter is not found.

        :param cfg: config dictionary. Can be a recursive dict (a dict of dicts, etc)
        :param path: Slash separated parameter path. E.g.: '/dict1/dict2/parameter'
    """
    try:
        return findConfigParameter(cfg, path)
    except KeyError:
        logger.warning("Could not find path in config: {!r}".format(path))
        return alt



def deleteParameter(cfg, parentPath, parName):
    """ Deletes a parameter from the config dict

        :param cfg: config dictionary
        :param parentPath: the path of the parent dict that contains the paremeter
        :param parName: name of the element that will be removed
    """
    logger.debug("Deleting {!r} from parent dict: {!r}".format(parName, parentPath))
    try:
        parentCfg = findConfigParameter(cfg, parentPath)
    except KeyError:
        logger.warning("Could not find path in configur: {!r}".format(parentPath))
        return

    try:
        del parentCfg[parName]
    except KeyError:
        logger.warning("Could not find {!r} in parent dict: {!r}".format(parName, parentPath))
        return
    except Exception as ex:
        # The element might not be a dict
        logger.warning("Error while deleting {!r} from parent dict {!r}: {}"
                       .format(parName, parentPath, ex))
        return



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Version and other info for this program
"""
from __future__ import print_function

import copy
import glob
import json
import logging
import os.path
import pprint
import shutil
import sys

from datetime import datetime

from argos.info import DEBUGGING, EXIT_CODE_SUCCESS, KEY_PROGRAM, PROJECT_NAME, VERSION, KEY_VERSION
from argos.inspector.registry import InspectorRegistry, DEFAULT_INSPECTOR
from argos.qt import QtCore, QtWidgets, QtSlot
from argos.qt.misc import handleException, initQApplication
from argos.reg.basereg import nameToIdentifier
from argos.repo.colors import CmLibSingleton, DEF_FAV_COLOR_MAPS
from argos.repo.registry import globalRtiRegistry
from argos.repo.repotreemodel import RepoTreeModel
from argos.utils.config import getConfigParameter, deleteParameter
from argos.utils.dirs import argosConfigDirectory, normRealPath, ensureFileExists
from argos.utils.moduleinfo import versionStrToTuple
from argos.widgets.mainwindow import MainWindow, UpdateReason

logger = logging.getLogger(__name__)


def _updateConfig(cfg):
    """ Updates the Argos config dict for the new version after an update.

        Best to make sure that a backup has been made before.
    """
    logger.debug("_updateConfig called")

    if not cfg: # The config may be empty, e.g. when the first time the application is started
        logger.debug("Empty config file in _updateConfig. Returning.")
        return cfg

    cfg = copy.deepcopy(cfg)

    # Version is written to config from 0.4.0, so we assume version <= 0.3.x if no version was found
    # We drop the postfix (e.g. 'rc1') from the version tuples so that version tuples can be
    # compared directly with logical operators.
    cfgVersion = versionStrToTuple(cfg.get(KEY_VERSION, '0.3.99'))[0:3]
    appVersion = versionStrToTuple(VERSION)[0:3]

    if cfgVersion == appVersion:
        logger.info("Config file version {} == app version {}. Will use config file as-is."
                    .format(cfgVersion, appVersion))
        return cfg
    elif cfgVersion > appVersion:
        logger.warning("Config file version {} > app version {}. Will use config file as-is."
                    .format(cfgVersion, appVersion))
        return cfg
    else:
        logger.info("Config file version {} < app version {}. Updating config file to new version."
                    .format(cfgVersion, appVersion))
        assert appVersion > cfgVersion, "Sanity check"

    if cfgVersion < (0, 4, 0):
        logger.info("Updating config to 0.4.0")

        # Remove the tree headers because we added the kind, element type, and summary columns.
        for key, windowsCfg in getConfigParameter(cfg, 'windows', alt={}).items():
            logger.debug("Updating windows config: {}".format(key))
            deleteParameter(windowsCfg, "layout/repoWidget", "treeHeaders")

    return cfg


_Q_APP = None # Keep reference to QApplication instance to prevent garbage collection

def qApplicationSingleton():
    """ Returns the QApplication object. Creates it if it doesn't exist.

        :rtype QtWidgets.QApplication:
    """
    global _Q_APP

    qApp = QtWidgets.QApplication.instance()
    if qApp is None:
        _Q_APP = qApp = QtWidgets.QApplication([])

    return qApp


class ArgosApplication(QtCore.QObject):
    """ The application singleton which holds global state.
    """
    def __init__(self, settingsFile=None, setExceptHook=True, runTestWalk=False):
        """ Constructor

            :param settingsFile: Config file from which the persistent settings are loaded.

            :param setExceptHook: Sets the global sys.except hook so that Qt shows a dialog box
                when an exception is raised.

                In debugging mode, the program will just quit in case of an exception. This is
                standard Python behavior but PyQt and PySide swallow exceptions by default (only a
                log message is displayed). The practice of swallowing exceptions fosters bad
                programming IHMO as it is easy to miss errors. I strongly recommend that you set
                the setExceptHook to True.

            :param runTestWalk: if True, all nodes are visited and the program exits.
        """
        super(ArgosApplication, self).__init__()

        self._runTestWalk = runTestWalk

        if not settingsFile:
            settingsFile = ArgosApplication.defaultSettingsFile()
            logger.debug("No config file specified. Using default: {}".format(settingsFile))

        self._settingsFile = ArgosApplication.userConfirmedSettingsFile(
            settingsFile, createWithoutConfirm=ArgosApplication.defaultSettingsFile())

        if setExceptHook:
            logger.debug("Setting sys.excepthook to Argos exception handling")
            sys.excepthook = handleException

        QtCore.qInstallMessageHandler(self.handleQtLogMessages)

        if DEBUGGING:
            self.qApplication.focusChanged.connect(self.focusChanged) # for debugging

        self._repo = RepoTreeModel()
        self._rtiRegistry = globalRtiRegistry()
        self._inspectorRegistry = InspectorRegistry()

        self._mainWindows = []
        self._settingsSaved = False  # boolean to prevent saving settings twice
        self._recentFiles = []    # list of recently opened files ([timeStampe, fileName] per file).
        self._maxRecentFiles = 10  # Maximum size of recent file

        #self.qApplication.lastWindowClosed.connect(self.quit)
        self.qApplication.aboutToQuit.connect(self.aboutToQuitHandler)

        # Activate-actions for all windows
        self.windowActionGroup = QtWidgets.QActionGroup(self)
        self.windowActionGroup.setExclusive(True)

        # Call setup when the event loop starts.
        QtCore.QTimer.singleShot(0, self.onEventLoopStarted)


    def onEventLoopStarted(self):
        """ Is called as soon as the event loop has started.
        """
        logger.debug("ArgosApplication.onEventLoopStarted called")

        # Raising all window because in OS-X window 0 is not shown.
        #self.raiseAllWindows()
        # activateWindow also solves the issue but doesn't work with the --inspector option.
        actions = self.windowActionGroup.actions()
        if actions:
            actions[0].trigger()

        if self._runTestWalk:
            testWalkDialog = self._mainWindows[0].testWalkDialog
            testWalkDialog.walkAllRepoNodes(allInspectors=True, allDetailTabs=True)
            allPassed = all(success for (success, _) in testWalkDialog.results)
            if not DEBUGGING:
                sys.exit(0 if allPassed else 1)


    @classmethod
    def defaultSettingsFile(cls):
        """ Returns the path to the default settings file
        """
        return normRealPath(os.path.join(argosConfigDirectory(), 'settings.json'))


    @classmethod
    def userConfirmedSettingsFile(cls, settingsFile, createWithoutConfirm=None):
        """ Asks the user to confirm creating the settings file if it does not exist and is not
            in the createWithoutConfirm list.

            If settingsFile is a relative path it will be concatenated to the argos config dir.

            :returns: The path of the setting file.
        """
        settingsFile = os.path.join(argosConfigDirectory(), settingsFile)
        logger.debug("Testing if settings file exists: {}".format(settingsFile))
        if os.path.exists(settingsFile):
            return settingsFile
        else:
            if settingsFile in createWithoutConfirm:
                return ensureFileExists(settingsFile)
            else:
                _app = qApplicationSingleton() # make sure QApplication exists.
                button = QtWidgets.QMessageBox.question(
                    None, "Create settings file?",
                    "The setting file cannot be found: {} \n\nCreate new config file?"
                                                        .format(settingsFile))

                if button == QtWidgets.QMessageBox.Yes:
                    return ensureFileExists(settingsFile)
                else:
                    logger.warning("No settings file created. Argos won't save persistent settings.")
                    return None


    ##############
    # Properties #
    ##############

    @property
    def qApplication(self):
        """ Returns the QApplication object. Equivalent to QtWidgets.qApp.

            :rtype QtWidgets.QApplication:
        """
        # TODO: replace the lines below by qApplicationSingleton()
        global _Q_APP

        qApp = QtWidgets.QApplication.instance()
        if qApp is None:
            logger.debug("Creating QApplication")
            _Q_APP = qApp = initQApplication()

        return qApp


    @property
    def repo(self):
        """ Returns the global repository
        """
        return self._repo


    @property
    def rtiRegistry(self):
        """ Returns the repository tree item (rti) registry
        """
        return self._rtiRegistry


    @property
    def inspectorRegistry(self):
        """ Returns the repository tree item (rti) registry
        """
        return self._inspectorRegistry


    @property
    def mainWindows(self):
        """ Returns the list of MainWindows. For read-only purposes only.
        """
        return self._mainWindows


    @property
    def settingsFile(self):
        """ Returns the name of the settings file
        """
        return self._settingsFile


    ###########
    # Methods #
    ###########

    def focusChanged(self, old, now):
        """ Is called when the focus changes. Useful for debugging.
        """
        #logger.debug("Focus changed from {} to {}".format(old, now))
        pass


    @classmethod
    def handleQtLogMessages(cls, qMsgType, context, msg):
        """ Forwards Qt log messages to the python log system.

            This ensures that they end up in application log files instead of just being printed to
            stderr at application exit.

            This function must be installed with QtCore.qInstallMessageHandler.
            See https://doc.qt.io/qt-5/qtglobal.html#qInstallMessageHandler
        """
        # Still print the message to stderr, just like Qt does by default.
        print(msg, file=sys.stderr, flush=True)

        if qMsgType == QtCore.QtDebugMsg:
            logger.debug(msg)
        elif qMsgType == QtCore.QtInfoMsg:
            logger.info(msg)
        elif qMsgType == QtCore.QtWarningMsg:
            logger.warning(msg)
        elif qMsgType == QtCore.QtCriticalMsg:
            logger.error(msg)
        elif qMsgType == QtCore.QtFatalMsg:
            logger.error(msg)
        else:
            logger.critical("Qt message of unknown type {}: {}".format(qMsgType, msg))


    def _makeConfigBackup(self, cfg):
        """ Creates a backup of the config file if the application version is not the same
            as the version from the config file.
        """
        if not cfg: # The config may be empty, e.g. when the first time the application is started
            return cfg

        cfgVersion = cfg.get(KEY_VERSION, '0.3.1') # version is present in config from 0.3.2 and up

        if cfgVersion != VERSION:
            logger.info("Config file version {} differs from application version {}."
                        .format(cfgVersion, VERSION))

            cfgBackup = "{}.v{}.backup".format(self._settingsFile, cfgVersion)
            logger.info("Making a backup to: {}".format(os.path.abspath(cfgBackup)))
            shutil.copy2(self._settingsFile, cfgBackup)

        return cfg


    def marshall(self):
        """ Returns a dictionary to save in the persistent settings
        """
        cfg = {}

        cfg[KEY_PROGRAM] = PROJECT_NAME
        cfg[KEY_VERSION] = VERSION

        cmLib = CmLibSingleton.instance()
        cfg['cmFavorites'] = [colorMap.key for colorMap in cmLib.color_maps
                     if colorMap.meta_data.favorite]
        logger.debug("cmFavorites: {}".format(cfg['cmFavorites']))

        cfg['recentFiles'] = self._recentFiles

        cfg['plugins'] = {}
        cfg['plugins']['inspectors'] = self.inspectorRegistry.marshall()
        cfg['plugins']['file-formats'] = self.rtiRegistry.marshall()

        # Save windows as a dict instead of a list to improve readability of the resulting JSON
        cfg['windows'] = {}
        for winNr, mainWindow in enumerate(self.mainWindows):
            key = "win-{:d}".format(winNr)
            cfg['windows'][key] = mainWindow.marshall()

        return cfg


    def unmarshall(self, cfg, inspectorFullName):
        """ Initializes itself from a config dict form the persistent settings.

            :param inspectorFullName: If not None a window with this inspector is created.
            If an inspector window with this inspector is created from the config file, this
            parameter is ignored.
        """
        cmLib = CmLibSingleton.instance()
        if not cmLib.color_maps:
            logger.warning("No color maps loaded yet. Favorites will be empty.")
            if DEBUGGING:
                assert False, "No color maps loaded yet. Favorites will be empty."

        favKeys = cfg.get('cmFavorites', DEF_FAV_COLOR_MAPS)
        for colorMap in cmLib.color_maps:
            colorMap.meta_data.favorite = colorMap.key in favKeys

        self._recentFiles = cfg.get('recentFiles', [])

        pluginCfg = cfg.get('plugins', {})

        self.inspectorRegistry.unmarshall(pluginCfg.get('inspectors', {}))
        self.rtiRegistry.unmarshall(pluginCfg.get('file-formats', {}))

        for winId, winCfg in cfg.get('windows', {}).items():
            assert winId.startswith('win-'), "Win ID doesn't start with 'win-': {}".format(winId)
            self.addNewMainWindow(cfg=winCfg)

        if inspectorFullName is not None:
            windows = [win for win in self._mainWindows
                       if win.inspectorFullName == inspectorFullName]
            if len(windows) == 0:
                logger.info("Creating window for inspector: {!r}".format(inspectorFullName))
                try:
                    win = self.addNewMainWindow(inspectorFullName=inspectorFullName)
                except KeyError:
                    logger.warning("No inspector found with ID: {}".format(inspectorFullName))
            else:
                for win in windows:
                    win.raise_()

        if len(self.mainWindows) == 0:
            logger.info("No open windows in settings or command line (creating one).")
            self.addNewMainWindow(inspectorFullName=DEFAULT_INSPECTOR)


    def saveSettings(self):
        """ Saves the persistent settings to file.
        """
        try:
            if not self._settingsFile:
                logger.info("No settings file specified. Not saving persistent state.")
            else:
                logger.info("Saving settings to: {}".format(self._settingsFile))
                settings = self.marshall()
                logger.debug("File formats: {}".format(settings['plugins']['file-formats']))
                try:
                    jsonStr = json.dumps(settings, sort_keys=True, indent=4)
                except Exception as ex:
                    logger.error("Failed to serialize settings to JSON: {}".format(ex))
                    logger.error("Settings: ...\n" + pprint.pformat(settings, width=100))
                    raise
                else:
                    with open(self._settingsFile, 'w') as settingsFile:
                        settingsFile.write(jsonStr)

        except Exception as ex:
            # Continue, even if saving the settings fails.
            logger.exception(ex)
            if DEBUGGING:
                raise
        finally:
            self._settingsSaved = True


    def loadSettings(self, inspectorFullName):
        """ Loads the settings from file and populates the application object from it.

            Will update the config (and make a backup of the config file) if the version number
            has changed.

            :param inspectorFullName: If not None, a window with this inspector is created.
                If an inspector window with this inspector is created from the config file, this
                parameter is ignored.
        """
        if not os.path.exists(self._settingsFile):
            logger.warning("Settings file does not exist: {}".format(self._settingsFile))

        try:
            with open(self._settingsFile, 'r') as settingsFile:
                jsonStr = settingsFile.read()

            if jsonStr:
                cfg = json.loads(jsonStr)
            else:
                cfg = {}
        except Exception as ex:
            logger.error("Error {} while reading settings file: {}"
                           .format(ex, self._settingsFile))
            raise # in case of a syntax error it's probably best to exit.

        self._makeConfigBackup(cfg)  # If the version has changed.
        cfg = _updateConfig(cfg)     # If the version has changed.

        self.unmarshall(cfg, inspectorFullName)  # Always call unmarshall.


    def saveSettingsIfLastWindow(self):
        """ Writes the persistent settings if this is the last window and the settings have not yet
            been saved.
        """
        if not self._settingsSaved and len(self.mainWindows) <= 1:
            self.saveSettings()


    def loadFiles(self, fileNames):
        """ Loads files into the repository as repo tree items of class rtiClass.
            Auto-detects using the extensions when rtiClass is None

            :param filePatterns: list of file names or unix like file expansions (globs),
                For example filePatterns = ['my_file.nc, 'your_file.nc']
                For example filePatterns = ['*.h5']
        """
        for fileName in fileNames:
            self.repo.loadFile(fileName, rtiRegItem=None)


    def getRecentFiles(self):
        """ Adds the files to the list of recently opened files
        """
        return self._recentFiles


    def addToRecentFiles(self, fileNames, rtiRegItemName):
        """ Adds the files to the list of recently opened files
        """
        # We only use the timestamp for sorting so we can store it as string. Easy to (un)marshall.
        timeStamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        # List of (fileName, rtiRegItemName) tuples to test if the new item already is in this list.
        recentItems = [(item[1], item[2]) for item in self._recentFiles]

        for fileName in fileNames:
            try:
                existingIndex = recentItems.index((fileName, rtiRegItemName))
            except ValueError:
                self._recentFiles.append([timeStamp, fileName, rtiRegItemName])
            else:
                # Only update timestamp if item is already in de list.
                self._recentFiles[existingIndex][0] = timeStamp

        self._recentFiles.sort(reverse=True)
        self._recentFiles = self._recentFiles[0:self._maxRecentFiles]


    def repopulateAllWindowMenus(self):
        """ Repopulates the Window menu of all main windows from scratch.

            To be called when a main window is created or removed.
        """
        for win in self.mainWindows:
            win.repopulateWindowMenu(self.windowActionGroup)


    @QtSlot()
    def addNewMainWindow(self, cfg=None, inspectorFullName=None):
        """ Creates and shows a new MainWindow.

            If inspectorFullName is set, it will set the identifier from that name.
            If the inspector identifier is not found in the registry, a KeyError is raised.
        """
        mainWindow = MainWindow(self)
        self.mainWindows.append(mainWindow)

        self.windowActionGroup.addAction(mainWindow.activateWindowAction)
        self.repopulateAllWindowMenus()

        if cfg:
            mainWindow.unmarshall(cfg)

        if inspectorFullName:
            inspectorId = nameToIdentifier(inspectorFullName)
            mainWindow.setInspectorById(inspectorId)

        if mainWindow.inspectorRegItem: # can be None at start
            inspectorId = mainWindow.inspectorRegItem.identifier
            mainWindow.getInspectorActionById(inspectorId).setChecked(True)
            logger.info("Created new window with inspector: {}"
                        .format(mainWindow.inspectorRegItem.name))
        else:
            logger.info("Created new window without inspector")

        mainWindow.drawInspectorContents(reason=UpdateReason.NEW_MAIN_WINDOW)
        mainWindow.show()

        if sys.platform.startswith('darwin'):
            # Calling raise before the QApplication.exec_ only shows the last window
            # that was added. Therefore we also call activeWindow. However, this may not
            # always be desirable. TODO: make optional?
            mainWindow.raise_()
            pass

        return mainWindow


    def removeMainWindow(self, mainWindow):
        """ Removes the mainWindow from the list of windows. Saves the settings
        """
        logger.debug("removeMainWindow called")

        self.windowActionGroup.removeAction(mainWindow.activateWindowAction)
        self.repopulateAllWindowMenus()

        self.mainWindows.remove(mainWindow)


    def raiseAllWindows(self):
        """ Raises all application windows.
        """
        logger.debug("raiseAllWindows called")
        for mainWindow in self.mainWindows:
            logger.debug("Raising {}".format(mainWindow._instanceNr))
            mainWindow.raise_()


    def exit(self, exitCode):
        """ Saves settings and exits the program with a certain exit code.
        """
        self.saveSettings()
        self.qApplication.closeAllWindows()
        self.qApplication.exit(exitCode)


    def quit(self):
        """ Saves settings and exits the program with exit code 0 (success).
        """
        self.exit(EXIT_CODE_SUCCESS)


    def aboutToQuitHandler(self):
        """ Called by Qt when the application is quitting
        """
        # No need to save the settings here as long as ArgosApplication.exit has been called.
        logger.info("Argos is about to quit")
        # QtWidgets.QMessageBox.warning(None, "Exit in progress", "We are quitting")


    def execute(self):
        """ Executes all main windows by starting the Qt main application
        """
        logger.info("Starting Argos event loop...")
        exitCode = self.qApplication.exec_()
        logger.info("Argos event loop finished with exit code: {}".format(exitCode))
        return exitCode




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Color selector functionality. E.g. for use in the plugin config dialog
"""
import logging
from argos.qt import QtCore, QtGui, QtWidgets, Qt
from argos.utils.defs import InvalidInputError

logger = logging.getLogger(__name__)

class ColorSelectWidget(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(ColorSelectWidget, self).__init__(parent=parent)

        self.mainLayout = QtWidgets.QHBoxLayout()
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.mainLayout)

        self.lineEditor = QtWidgets.QLineEdit()
        self.lineEditor.setToolTip("Color hex code: a '#' followed by 6 hex-digits.")
        self.mainLayout.addWidget(self.lineEditor)
        self.setFocusProxy(self.lineEditor)

        regExp = QtCore.QRegExp(r'#[0-9A-F]{6}', Qt.CaseInsensitive)
        validator = QtGui.QRegExpValidator(regExp, parent=self)
        self.lineEditor.setValidator(validator)

        self.pickButton = QtWidgets.QPushButton()
        self.pickButton.setText("Color Picker...")
        self.pickButton.setToolTip("Open color dialog.")
        self.pickButton.setFocusPolicy(Qt.NoFocus)
        self.pickButton.clicked.connect(self.openColorDialog)

        self.mainLayout.addWidget(self.pickButton)


    def setData(self, qColor):
        """ Provides the main editor widget with a data to manipulate.
        """
        self.lineEditor.setText(qColor.name().upper())
        self.lineEditor.setFocus(Qt.OtherFocusReason)  #


    def getData(self):
        """ Gets data from the editor widget.
        """
        text = self.lineEditor.text()
        if not text.startswith('#'):
            text = '#' + text

        validator = self.lineEditor.validator()
        if validator is not None:
            state, text, _ = validator.validate(text, 0)
            if state != QtGui.QValidator.Acceptable:
                raise InvalidInputError("Invalid input: {!r}".format(text))

        return  text


    def openColorDialog(self):
        """ Opens a QColorDialog for the user
        """
        try:
            currentColor = self.getData()
        except InvalidInputError:
            currentColor = "#FFFFFF"

        qColor = QtWidgets.QColorDialog.getColor(QtGui.QColor(currentColor), self)

        if qColor.isValid():
            self.setData(qColor)



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Version and other info for this program
"""
import os, sys

# We bypass the argparse mechanism in main.py because this import is executed before main.main()
DEBUGGING = ('-d' in sys.argv or '--debug' in sys.argv)
TESTING = True # add some test menu options
PROFILING = False# and DEBUGGING

VERSION = '0.4.3'
REPO_NAME = "argos"
SCRIPT_NAME = "argos"
PACKAGE_NAME = "argos"
PROJECT_NAME = "Argos"
SHORT_DESCRIPTION = "Argos HDF5/NetCDF/scientific data viewer."
PROJECT_URL = "https://github.com/titusjan/argos"
AUTHOR = "Pepijn Kenter"
EMAIL = "titusjan@gmail.com"
ORGANIZATION_NAME = "titusjan"
ORGANIZATION_DOMAIN = "titusjan.nl"

EXIT_CODE_SUCCESS = 0
EXIT_CODE_ERROR = 1
EXIT_CODE_COMMAND_ARGS = 2
EXIT_CODE_RESTART = 66 # Indicates the program is being 'restarted'

KEY_PROGRAM = '_program'
KEY_VERSION = '_version'

# TODO: move to utils locations
def program_directory():
    """ Returns the program directory where this program is installed
    """
    return os.path.abspath(os.path.dirname(__file__))


def resource_directory():
    """ Returns directory with resources (images, style sheets, etc)
    """
    return os.path.join(program_directory(), 'img/')



def icons_directory():
    """ Returns the icons directory
    """
    return os.path.join(program_directory(), 'img/snipicons')




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Contains the CTIs that represent Qt data types.
"""
import logging

from argos.config.abstractcti import AbstractCti, AbstractCtiEditor
from argos.config.boolcti import BoolCti
from argos.config.choicecti import ChoiceCti
from argos.config.intcti import IntCti
from argos.config.floatcti import FloatCti
from argos.info import DEBUGGING
from argos.qt import Qt, QtCore, QtGui, QtWidgets, QtSlot
from argos.utils.cls import check_is_a_string
from argos.utils.defs import InvalidInputError



logger = logging.getLogger(__name__)


PEN_STYLE_DISPLAY_VALUES = ('solid', 'dashed', 'dotted', 'dash-dot', 'dash-dot-dot')
PEN_STYLE_CONFIG_VALUES = (Qt.SolidLine, Qt.DashLine, Qt.DotLine,
                           Qt.DashDotLine, Qt.DashDotDotLine)


def createPenStyleCti(nodeName, defaultData=0, includeNone=False):
    """ Creates a ChoiceCti with Qt PenStyles.
        If includeEmtpy is True, the first option will be None.
    """
    displayValues=PEN_STYLE_DISPLAY_VALUES
    configValues=PEN_STYLE_CONFIG_VALUES
    if includeNone:
        displayValues = [''] + list(displayValues)
        configValues = [None] + list(configValues)
    return ChoiceCti(nodeName, defaultData,
                     displayValues=displayValues, configValues=configValues)


def createPenWidthCti(nodeName, defaultData=1.0, zeroValueText=None):
    """ Creates a FloatCti with defaults for configuring a QPen width.

        If specialValueZero is set, this string will be displayed when 0.0 is selected.
        If specialValueZero is None, the minValue will be 0.1
    """
    # A pen line width of zero indicates a cosmetic pen. This means that the pen width is
    # always drawn one pixel wide, independent of the transformation set on the painter.
    # Note that line widths other than 1 may be slow when anti aliasing is on.
    return FloatCti(nodeName, defaultData=defaultData, specialValueText=zeroValueText,
                    minValue=0.1 if zeroValueText is None else 0.0,
                    maxValue=100, stepSize=0.1, decimals=1)


def fontFamilyIndex(qFont, families):
    """ Searches the index of qFont.family in the families list.
        If qFont.family() is not in the list, the index of qFont.defaultFamily() is returned.
        If that is also not present an error is raised.
    """
    try:
        return families.index(qFont.family())
    except ValueError:
        if False and DEBUGGING:
            raise
        else:
            logger.warning("{} not found in font families, using default.".format(qFont.family()))
            return families.index(qFont.defaultFamily())


def fontWeightIndex(qFont, weights):
    """ Searches the index of qFont.family in the weight list.
        If qFont.weight() is not in the list, the index of QFont.Normal is returned.
        If that is also not present an error is raised.
    """
    try:
        return weights.index(qFont.weight())
    except ValueError:
        if False and DEBUGGING:
            raise
        else:
            logger.warning("{} not found in font weights, using normal.".format(qFont.weight()))
            return weights.index(QtGui.QFont.Normal)


class ColorCti(AbstractCti):
    """ Config Tree Item to store a color.
    """
    def __init__(self, nodeName, defaultData=''):
        """ Constructor.
            For the (other) parameters see the AbstractCti constructor documentation.
        """
        super(ColorCti, self).__init__(nodeName, defaultData)

    def _enforceDataType(self, data):
        """ Converts to str so that this CTI always stores that type.
        """
        qColor = QtGui.QColor(data)
        if not qColor.isValid():
            raise ValueError("Invalid color specification: {!r}".format(data))
        return qColor

    def _dataToString(self, data):
        """ Conversion function used to convert the (default)data to the display value.
        """
        return data.name().upper()

    @property
    def decoration(self):
        """ Returns the data (QColor) to be displayed as decoration
        """
        return self.data


    def _nodeMarshall(self):
        """ Returns the non-recursive marshalled value of this CTI. Is called by marshall()
        """
        return self.data.name()


    def _nodeUnmarshall(self, data):
        """ Initializes itself non-recursively from data. Is called by unmarshall()
        """
        self.data = QtGui.QColor(data)


    def createEditor(self, delegate, parent, option):
        """ Creates a ColorCtiEditor.
            For the parameters see the AbstractCti constructor documentation.
        """
        return ColorCtiEditor(self, delegate, parent=parent)



class ColorCtiEditor(AbstractCtiEditor):
    """ A CtiEditor which contains a QLineEdit for editing ColorCti objects.
    """
    def __init__(self, cti, delegate, parent=None):
        """ See the AbstractCtiEditor for more info on the parameters
        """
        super(ColorCtiEditor, self).__init__(cti, delegate, parent=parent)

        lineEditor = QtWidgets.QLineEdit(parent)
        regExp = QtCore.QRegExp(r'#?[0-9A-F]{6}', Qt.CaseInsensitive)
        validator = QtGui.QRegExpValidator(regExp, parent=lineEditor)
        lineEditor.setValidator(validator)

        self.lineEditor = self.addSubEditor(lineEditor, isFocusProxy=True)

        pickButton = QtWidgets.QToolButton()
        pickButton.setText("...")
        pickButton.setToolTip("Open color dialog.")
        pickButton.setFocusPolicy(Qt.NoFocus)
        pickButton.clicked.connect(self.openColorDialog)

        self.pickButton = self.addSubEditor(pickButton)


    def finalize(self):
        """ Is called when the editor is closed. Disconnect signals.
        """
        self.pickButton.clicked.disconnect(self.openColorDialog)
        super(ColorCtiEditor, self).finalize()


    def openColorDialog(self):
        """ Opens a QColorDialog for the user
        """
        try:
            currentColor = self.getData()
        except InvalidInputError:
            currentColor = self.cti.data

        qColor = QtWidgets.QColorDialog.getColor(currentColor, self)

        if qColor.isValid():
            self.setData(qColor)
            self.commitAndClose()


    def setData(self, qColor):
        """ Provides the main editor widget with a data to manipulate.
        """
        self.lineEditor.setText(qColor.name().upper())


    def getData(self):
        """ Gets data from the editor widget.
        """
        text = self.lineEditor.text()
        if not text.startswith('#'):
            text = '#' + text

        validator = self.lineEditor.validator()
        if validator is not None:
            state, text, _ = validator.validate(text, 0)
            if state != QtGui.QValidator.Acceptable:
                raise InvalidInputError("Invalid input: {!r}".format(text))

        return  QtGui.QColor(text)


class FontCti(AbstractCti):
    """ Config Tree Item that configures a QFont

        The target font will be be updated by calling updateTarget (of the root CTI)
    """
    def __init__(self, targetWidget, nodeName='font', defaultData=None):
        """ Constructor.

            :param targetWidget: a QWidget that must have a setFont() method
            For the (other) parameters see the AbstractCti constructor documentation.
        """
        super(FontCti, self).__init__(nodeName,
                                defaultData=QtGui.QFont() if defaultData is None else defaultData)

        self._targetWidget = targetWidget

        self.familyCti = self.insertChild(
            FontChoiceCti("family", defaultFamily=self.defaultData.family()))

        self.pointSizeCti = self.insertChild(
            IntCti("size", self.defaultData.pointSize(),
                   minValue=1, maxValue=500, stepSize=1, suffix=' pt'))

        QtF = QtGui.QFont
        weights = [QtF.Light, QtF.Normal, QtF.DemiBold, QtF.Bold, QtF.Black]
        self.weightCti = self.insertChild(
            ChoiceCti("weight", defaultData=fontWeightIndex(self.defaultData, weights),
                      displayValues=['Light', 'Normal', 'DemiBold', 'Bold', 'Black'],
                      configValues=weights))

        self.italicCti = self.insertChild(BoolCti("italic", self.defaultData.italic()))


    def _enforceDataType(self, data):
        """ Converts to str so that this CTI always stores that type.
        """
        result = QtGui.QFont(data)
        return result


    @property
    def data(self):
        """ Returns the font of this item.
        """
        return self._data


    @data.setter
    def data(self, data):
        """ Sets the font data of this item.
            Does type conversion to ensure data is always of the correct type.

            Also updates the children (which is the reason for this property to be overloaded.
        """
        self._data = self._enforceDataType(data) # Enforce self._data to be a QFont
        self.familyCti.data = fontFamilyIndex(self.data, list(self.familyCti.iterConfigValues))
        self.pointSizeCti.data = self.data.pointSize()
        self.weightCti.data = fontWeightIndex(self.data, list(self.weightCti.iterConfigValues))
        self.italicCti.data = self.data.italic()


    @property
    def defaultData(self):
        """ Returns the default font of this item.
        """
        return self._defaultData


    @defaultData.setter
    def defaultData(self, defaultData):
        """ Sets the data of this item.
            Does type conversion to ensure default data is always of the correct type.
        """
        self._defaultData = self._enforceDataType(defaultData) # Enforce to be a QFont
        self.familyCti.defaultData = fontFamilyIndex(self.defaultData,
                                                     list(self.familyCti.iterConfigValues))
        self.pointSizeCti.defaultData = self.defaultData.pointSize()
        self.weightCti.defaultData = self.defaultData.weight()
        self.italicCti.defaultData = self.defaultData.italic()


    @property
    def displayValue(self):
        """ Returns the string representation of data for use in the tree view.
            Returns the empty string.
        """
        return ''


    @property
    def debugInfo(self):
        """ Returns the string representation of data for use in the tree view.
            Returns the empty string. If info.DEBUGGING is True, QFont.toString is used.
        """
        return self._dataToString(self.data)


    def _dataToString(self, data):
        """ Conversion function used to convert the (default)data to the display value.
        """
        return data.toString() # data is a QFont


    def _updateTargetFromNode(self):
        """ Applies the font config settings to the target widget's font.

            That is the targetWidget.setFont() is called with a font create from the config values.
        """
        font = self.data
        if self.familyCti.configValue:
            font.setFamily(self.familyCti.configValue)
        else:
            font.setFamily(QtGui.QFont().family()) # default family
        font.setPointSize(self.pointSizeCti.configValue)
        font.setWeight(self.weightCti.configValue)
        font.setItalic(self.italicCti.configValue)
        self._targetWidget.setFont(font)


    def _nodeGetNonDefaultsDict(self):
        """ Retrieves this nodes` values as a dictionary to be used for persistence.
            Non-recursive auxiliary function for getNonDefaultsDict
        """
        dct = {}
        if self.data != self.defaultData:
            dct['data'] = self.data.toString() # calls QFont.toString()
        return dct


    def _nodeSetValuesFromDict(self, dct):
        """ Sets values from a dictionary in the current node.
            Non-recursive auxiliary function for setValuesFromDict
        """
        if 'data' in dct:
            qFont = QtGui.QFont()
            success = qFont.fromString(dct['data'])
            if not success:
                msg = "Unable to create QFont from string {!r}".format(dct['data'])
                logger.warning(msg)
                if DEBUGGING:
                    raise ValueError(msg)
            self.data = qFont


    def _nodeMarshall(self):
        """ Returns the non-recursive marshalled value of this CTI. Is called by marshall()
        """
        return self.data.toString() # calls QFont.toString()


    def _nodeUnmarshall(self, data):
        """ Initializes itself non-recursively from data. Is called by unmarshall()
        """
        qFont = QtGui.QFont()
        success = qFont.fromString(data)
        if not success:
            msg = "Unable to create QFont from string {!r}".format(data)
            logger.warning(msg)
            if DEBUGGING:
                raise ValueError(msg)
        self.data = qFont



    def createEditor(self, delegate, parent, option):
        """ Creates a FontCtiEditor.
            For the parameters see the AbstractCti documentation.
        """
        return FontCtiEditor(self, delegate, parent=parent)



class FontCtiEditor(AbstractCtiEditor):
    """ A CtiEditor which contains a label and a button for opening the Qt font dialog
    """
    def __init__(self, cti, delegate, parent=None):
        """ See the AbstractCtiEditor for more info on the parameters
        """
        super(FontCtiEditor, self).__init__(cti, delegate, parent=parent)

        pickButton = QtWidgets.QToolButton()
        pickButton.setText("...")
        pickButton.setToolTip("Open font dialog.")
        pickButton.setFocusPolicy(Qt.NoFocus)
        pickButton.clicked.connect(self.execFontDialog)

        self.pickButton = self.addSubEditor(pickButton)

        if DEBUGGING:
            self.label = self.addSubEditor(QtWidgets.QLabel(self.cti.displayValue))
            labelSizePolicy = self.label.sizePolicy()
            self.label.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding,
                                     labelSizePolicy.verticalPolicy())


    def finalize(self):
        """ Is called when the editor is closed. Disconnect signals.
        """
        self.pickButton.clicked.disconnect(self.execFontDialog)
        super(FontCtiEditor, self).finalize()


    def execFontDialog(self):
        """ Opens a QColorDialog for the user
        """
        currentFont = self.getData()
        newFont, ok = QtWidgets.QFontDialog.getFont(currentFont, self)
        if ok:
            self.setData(newFont)
        else:
            self.setData(currentFont)
        self.commitAndClose()


    def setData(self, data):
        """ Provides the main editor widget with a data to manipulate.
        """
        # Set the data in the 'editor_data' property of the pickButton to that getData
        # can pass the same value back into the CTI.
        self.pickButton.setProperty("editor_data", data)


    def getData(self):
        """ Gets data from the editor widget.
        """
        return self.pickButton.property("editor_data")




class FontChoiceCti(ChoiceCti):
    """ A ChoiceCti that allows selecting one of the installed fonts families with a QFontCombobox

        The configValues are determined automatically, they cannot be set in the constructor.

        The QFontCombobox is not editable. I could get it to work well with the FontCti and
        automatic determination of the configValue. It doesn't add much anyway IHMO.
    """
    def __init__(self, nodeName, defaultFamily='Helvetica'):
        """ Constructor.

            :param defaultFamily: A string representing the defaultValue.
            :editable: True if the underlying QFontComboBox is editable. The default is False as
                it does not work well with the FontCti.

            For the (other) parameters see the AbstractCti constructor documentation.
        """
        check_is_a_string(defaultFamily)

        # Get a list of of configValues by reading them from a temporary QFontComboBox.
        tempFontComboBox = QtWidgets.QFontComboBox()
        configValues = []
        defaultData = 0
        for idx in range(tempFontComboBox.count()):
            fontFamily = tempFontComboBox.itemText(idx)
            configValues.append(fontFamily)
            if fontFamily.lower() == defaultFamily.lower():
                defaultData = idx

        # Set after self._displayValues are defined. The parent constructor calls _enforceDataType
        super(FontChoiceCti, self).__init__(nodeName, defaultData, configValues=configValues)


    def createEditor(self, delegate, parent, option):
        """ Creates a ChoiceCtiEditor.
            For the parameters see the AbstractCti constructor documentation.
        """
        return FontChoiceCtiEditor(self, delegate, parent=parent)



class FontChoiceCtiEditor(AbstractCtiEditor):
    """ A CtiEditor which contains a QCombobox for editing ChoiceCti objects.
    """
    def __init__(self, cti, delegate, parent=None):
        """ See the AbstractCtiEditor for more info on the parameters
        """
        super(FontChoiceCtiEditor, self).__init__(cti, delegate, parent=parent)

        comboBox = QtWidgets.QFontComboBox()
        # The QFontCombobox is not editable. because an non-existing value resulted in a QFont()
        # without parameter whose font family was probably some style sheet value. This didn't
        # work well with the extra options of the automatic configValue generation and with the
        # a possible parent FontCti: a non-exisiting family yields a QFont() with a non-existing
        # family, proably a style-sheet name, which I don't know how to handle.
        comboBox.setEditable(False)

        # The current font family is not properly selected when the combobox is created (at least
        # on OS-X). Setting the setMaxVisibleItems didn't help.
        # http://stackoverflow.com/questions/11252299/pyqt-qcombobox-setting-number-of-visible-items-in-dropdown
        # comboBox.setMaxVisibleItems(15)
        # comboBox.setStyleSheet("QComboBox { combobox-popup: 0; }");

        comboBox.activated.connect(self.comboBoxActivated)
        self.comboBox = self.addSubEditor(comboBox, isFocusProxy=True)


    def finalize(self):
        """ Is called when the editor is closed. Disconnect signals.
        """
        self.comboBox.activated.disconnect(self.comboBoxActivated)
        super(FontChoiceCtiEditor, self).finalize()


    def setData(self, data):
        """ Provides the main editor widget with a data to manipulate.
        """
        self.comboBox.setCurrentIndex(data)


    def getData(self):
        """ Gets data from the editor widget.
        """
        return self.comboBox.currentIndex()


    @QtSlot(int)
    def comboBoxActivated(self, index):
        """ Is called when the user chooses an item in the combo box. The item's index is passed.
            Note that this signal is sent even when the choice is not changed.
        """
        self.delegate.commitData.emit(self)



class PenCti(BoolCti):
    """ Config Tree Item to configure a QPen for drawing lines.

        It will create children for the pen color, width and style. It will not create a child
        for the brush.
    """
    def __init__(self, nodeName, defaultData, resetTo=None, expanded=True,
                 includeNoneStyle=False, includeZeroWidth=False):
        """ Sets the children's default value using the resetTo value.

            The resetTo value must be a QPen or value that can be converted to QPen. It is used
            to initialize the child nodes' defaultValues. If resetTo is None, the default QPen
            will be used, which is a black solid pen of width 1.0.

            (resetTo is not called 'defaultData' since the PenCti itself always has a data and
            defaultData of None. That is, it does not store the data itself but relies on its
            child nodes). The default data, is used to indicate if the pen is enabled.

            If includeNonStyle is True, an None-option will be prepended to the style choice
        """
        super(PenCti, self).__init__(nodeName, defaultData, expanded=expanded,
                                     childrenDisabledValue=False)
        # We don't need a similar initFrom parameter.
        qPen = QtGui.QPen(resetTo)

        self.colorCti = self.insertChild(ColorCti('color', defaultData=qPen.color()))
        defaultIndex = PEN_STYLE_CONFIG_VALUES.index(qPen.style()) + int(includeNoneStyle)
        self.styleCti = self.insertChild(createPenStyleCti('style', defaultData=defaultIndex,
                                                           includeNone=includeNoneStyle))
        self.widthCti = self.insertChild(createPenWidthCti('width', defaultData=qPen.widthF(),
                                                zeroValueText=' ' if includeZeroWidth else None))


    @property
    def configValue(self):
        """ Creates a QPen made of the children's config values.
        """
        if not self.data:
            return None
        else:
            pen = QtGui.QPen()
            pen.setCosmetic(True)
            pen.setColor(self.colorCti.configValue)
            style = self.styleCti.configValue
            if style is not None:
                pen.setStyle(style)
            pen.setWidthF(self.widthCti.configValue)
            return pen


    def createPen(self, altStyle=None, altWidth=None):
        """ Creates a pen from the config values with the style overridden by altStyle if the
            None-option is selected in the combo box.
        """
        pen = self.configValue
        if pen is not None:

            style = self.findByNodePath('style').configValue
            if style is None and altStyle is not None:
                pen.setStyle(altStyle)

            width = self.findByNodePath('width').configValue
            if width == 0.0 and altWidth is not None:
                #logger.debug("Setting altWidth = {!r}".format(altWidth))
                pen.setWidthF(altWidth)

        return pen




